import keras as ks
from kgcnn.layers.scale import get as get_scaler
from ._model import model_disjoint
from kgcnn.layers.modules import Input
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from kgcnn.models.utils import update_model_kwargs
from keras.backend import backend as backend_to_use

# To be updated if model is changed in a significant way.
__model_version__ = "2023-09-07"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]
if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'HamNet' is not supported." % backend_to_use())

# Implementation of HamNet in `keras` from paper:
# HamNet: Conformation-Guided Molecular Representation with Hamiltonian Neural Networks
# by Ziyao Li, Shuwen Yang, Guojie Song, Lingsheng Cai
# Link to paper: https://arxiv.org/abs/2105.03688
# Original implementation: https://github.com/PKUterran/HamNet
# Later implementation: https://github.com/PKUterran/MoleculeClub
# Note: the 2. implementation is cleaner than the original code and has been used as template.


model_default = {
    "name": "HamNet",
    "inputs": [
        {'shape': (None,), 'name': "node_number", 'dtype': 'int64'},
        {'shape': (None, 3), 'name': "node_coordinates", 'dtype': 'float32'},
        {'shape': (None, 64), 'name': "edge_attributes", 'dtype': 'float32'},
        {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64'},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {"input_dim": 95, "output_dim": 64},
    "input_edge_embedding": {"input_dim": 5, "output_dim": 64},
    "message_kwargs": {"units": 128, "units_edge": 128},
    "fingerprint_kwargs": {"units": 128, "units_attend": 128, "depth": 2},
    "gru_kwargs": {"units": 128},
    "verbose": 10,
    "depth": 1,
    "union_type_node": "gru",
    "union_type_edge": "None",
    "given_coordinates": True,
    "output_embedding": "graph",
    "output_tensor_type": "padded",
    "output_to_tensor": None,  # deprecated
    'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
                   "activation": ['relu', 'relu', 'linear']},
    "output_scaling": None
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(name: str = None,
               inputs: list = None,
               input_tensor_type: str = None,
               cast_disjoint_kwargs: dict = None,
               input_embedding: dict = None,
               input_node_embedding: dict = None,
               input_edge_embedding: dict = None,
               verbose: int = None,
               message_kwargs: dict = None,
               gru_kwargs: dict = None,
               fingerprint_kwargs: dict = None,
               union_type_node: str = None,
               union_type_edge: str = None,
               given_coordinates: bool = None,
               depth: int = None,
               output_embedding: str = None,
               output_to_tensor: bool = None,
               output_mlp: dict = None,
               output_tensor_type: str = None,
               output_scaling: dict = None
               ):
    r"""Make `HamNet <https://arxiv.org/abs/2105.03688>`__ graph model via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.HamNet.model_default` .

    .. note::

        At the moment only the Fingerprint Generator for graph embeddings is implemented and coordinates must
        be provided as model input.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, coordinates, edges, edge_indices, ...]` with `given_coordinates` and
    with '...' indicating mask or ID tensors following the template below.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        name (str): Name of the model.
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        input_node_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        input_edge_embedding (dict): Dictionary of embedding arguments for edges unpacked in :obj:`Embedding` layers.
        verbose (int): Level of verbosity. For logging and printing.
        message_kwargs (dict): Dictionary of layer arguments unpacked in message passing layer for node updates.
        gru_kwargs (dict): Dictionary of layer arguments unpacked in gated recurrent unit update layer.
        fingerprint_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`HamNetFingerprintGenerator` layer.
        given_coordinates (bool): Whether coordinates are provided as model input, or are computed by the Model.
        union_type_edge (str): Union type of edge updates. Choose "gru", "naive" or "None".
        union_type_node (str): Union type of node updates. Choose "gru", "naive" or "None".
        depth (int): Depth or number of (message passing) layers of the model.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): Whether to cast model output to :obj:`tf.Tensor`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    di_inputs = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 0, 1, 1],
        index_assignment=[None, None, None, 0]
    )

    n, x, ed, disjoint_indices, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = di_inputs

    # Wrapping disjoint model.
    out = model_disjoint(
        [n, x, ed, disjoint_indices, batch_id_node, count_nodes],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        use_edge_embedding=("int" in inputs[2]['dtype']) if input_edge_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        input_edge_embedding=input_edge_embedding,
        given_coordinates=given_coordinates,
        gru_kwargs=gru_kwargs,
        message_kwargs=message_kwargs,
        fingerprint_kwargs=fingerprint_kwargs,
        output_embedding=output_embedding,
        output_mlp=output_mlp,
        union_type_edge=union_type_edge,
        union_type_node=union_type_node,
        depth=depth
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding,
        output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)
    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



import numpy as np
import jax.numpy as jnp
from kgcnn import __safe_scatter_max_min_to_zero__ as global_safe_scatter_max_min_to_zero


class binfo:
  kind: str = "b"
  bits: int = 8  # May be different for jax.
  min: bool = False
  max: bool = True
  dtype: np.dtype = "bool"


def dtype_infos(dtype):
    if dtype.kind in ["f", "c"]:
        return jnp.finfo(dtype)
    elif dtype.kind in ["i", "u"]:
        return jnp.iinfo(dtype)
    elif dtype.kind in ["b"]:
        return binfo()
    else:
        raise TypeError("Unknown dtype '%s' to get type info." % dtype)


def scatter_reduce_sum(indices, values, shape):
    zeros = jnp.zeros(shape, values.dtype)
    return zeros.at[indices].add(values)


def scatter_reduce_min(indices, values, shape):
    max_of_dtype = dtype_infos(values.dtype).max
    zeros = jnp.full(shape, max_of_dtype, values.dtype)
    out = zeros.at[indices].min(values)
    if global_safe_scatter_max_min_to_zero:
        has_scattered = jnp.zeros(shape, "bool")
        has_scattered = has_scattered.at[indices].set(jnp.ones_like(values, dtype="bool"))
        out = jnp.where(has_scattered, out, jnp.zeros_like(out))
    return out


def scatter_reduce_max(indices, values, shape):
    min_of_dtype = dtype_infos(values.dtype).min
    zeros = jnp.full(shape, min_of_dtype, values.dtype)
    out = zeros.at[indices].max(values)
    if global_safe_scatter_max_min_to_zero:
        has_scattered = jnp.zeros(shape, "bool")
        has_scattered = has_scattered.at[indices].set(jnp.ones_like(values, dtype="bool"))
        out = jnp.where(has_scattered, out, jnp.zeros_like(out))
    return out


def scatter_reduce_mean(indices, values, shape):
    zeros = jnp.zeros(shape, values.dtype)
    counts = jnp.zeros(shape, values.dtype)
    counts = counts.at[indices].add(jnp.ones_like(values))
    inverse_counts = jnp.nan_to_num(jnp.reciprocal(counts), posinf=0.0, neginf=0.0, nan=0.0)
    return zeros.at[indices].add(values)*inverse_counts


def scatter_reduce_softmax(indices, values, shape, normalize: bool = False):

    if normalize:
        zeros_min = jnp.zeros(shape, values.dtype)  # Zero is okay here
        data_segment_max = zeros_min.at[indices].max(values)
        data_max = jnp.take(data_segment_max, indices, axis=0)
        values = values - data_max

    values_exp = jnp.exp(values)
    zeros = jnp.zeros(shape, values.dtype)
    values_exp_sum = zeros.at[indices].add(values_exp)
    values_exp_sum = jnp.take(values_exp_sum, indices, axis=0)
    return values_exp / values_exp_sum


def repeat_static_length(x, repeats, axis=None, total_repeat_length: int = None):
    return jnp.repeat(x, repeats=repeats, axis=axis, total_repeat_length=total_repeat_length)


def decompose_ragged_tensor(x):
    raise NotImplementedError("Operation not supported by this backend '%s'." % __name__)


def norm(x, ord='fro', axis=None, keepdims=False):
    return jnp.norm(x, ord=ord, axis=axis, keepdims=keepdims)


def cross(x1, x2):
    return jnp.cross(x1, x2, axis=-1)



from keras.layers import Dense
from kgcnn.layers.conv import GCN
from kgcnn.layers.mlp import MLP, GraphMLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.pooling import PoolingNodes, PoolingWeightedNodes

# from keras_core.layers import Activation
# from kgcnn.layers.aggr import AggregateWeightedLocalEdges
# from kgcnn.layers.gather import GatherNodesOutgoing


def model_disjoint(inputs,
                   use_node_embedding: bool = None,
                   use_edge_embedding: bool = None,
                   input_node_embedding: dict = None,
                   input_edge_embedding: dict = None,
                   depth: int = None,
                   gcn_args: dict = None,
                   node_pooling_args: dict = None,
                   output_embedding: str = None,
                   output_mlp: dict = None,
                   ):
    n, e, disjoint_indices, batch_id_node, count_nodes = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        e = Embedding(**input_edge_embedding)(e)

    # Model
    n = Dense(gcn_args["units"], use_bias=True, activation='linear')(n)  # Map to units
    for i in range(0, depth):
        n = GCN(**gcn_args)([n, e, disjoint_indices])

        # # Equivalent as:
        # no = Dense(gcn_args["units"], activation="linear")(n)
        # no = GatherNodesOutgoing()([no, disjoint_indices])
        # nu = AggregateWeightedLocalEdges()([n, no, disjoint_indices, e])
        # n = Activation(gcn_args["activation"])(nu)

    # Output embedding choice
    if output_embedding == "graph":
        out = PoolingNodes(**node_pooling_args)([count_nodes, n, batch_id_node])  # will return tensor
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for `GCN` .")

    return out


def model_disjoint_weighted(inputs,
                            use_node_embedding: bool = None,
                            use_edge_embedding: bool = None,
                            input_node_embedding: dict = None,
                            input_edge_embedding: dict = None,
                            depth: int = None,
                            gcn_args: dict = None,
                            node_pooling_args: dict = None,
                            output_embedding: str = None,
                            output_mlp: dict = None,
                            ):
    n, nw, e, disjoint_indices, batch_id_node, count_nodes = inputs
    # Embedding, if no feature dimension

    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        e = Embedding(**input_edge_embedding)(e)

    # Model
    n = Dense(gcn_args["units"], use_bias=True, activation='linear')(n)  # Map to units
    for i in range(0, depth):
        n = GCN(**gcn_args)([n, e, disjoint_indices])

    # Output embedding choice
    if output_embedding == "graph":
        out = PoolingWeightedNodes(**node_pooling_args)([count_nodes, n, nw, batch_id_node])  # will return tensor
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for `GCN`")

    return out



import keras as ks
from kgcnn.ops.activ import *
from kgcnn.layers.attention import AttentiveHeadFP
from kgcnn.layers.mlp import MLP, GraphMLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.update import GRUUpdate
from kgcnn.layers.pooling import PoolingNodesAttentive


def model_disjoint(
        inputs,
        use_node_embedding: bool = None,
        use_edge_embedding: bool = None,
        input_node_embedding: dict = None,
        input_edge_embedding: dict = None,
        depthmol: int = None,
        depthato: int = None,
        dropout: float = None,
        attention_args: dict = None,
        output_embedding: str = None,
        output_mlp: dict = None
):
    n, ed, edi, batch_id_node, count_nodes = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)

    # Model
    nk = ks.layers.Dense(units=attention_args['units'])(n)
    ck = AttentiveHeadFP(use_edge_features=True, **attention_args)([nk, ed, edi])
    nk = GRUUpdate(units=attention_args['units'])([nk, ck])

    for i in range(1, depthato):
        ck = AttentiveHeadFP(**attention_args)([nk, ed, edi])
        nk = GRUUpdate(units=attention_args['units'])([nk, ck])
        nk = ks.layers.Dropout(rate=dropout)(nk)
    n = nk

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodesAttentive(units=attention_args['units'], depth=depthmol)([count_nodes, n, batch_id_node])
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported graph embedding for mode `AttentiveFP`")

    return out


from ._make import make_model, model_default
from ._make import make_model_edge, model_default_edge

__all__ = [
    "make_model",
    "model_default",
    "make_model_edge",
    "model_default_edge"
]



import keras as ks
from keras import ops
from keras.layers import Dense, Multiply, Add, Layer
from kgcnn.layers.aggr import AggregateLocalEdges
from kgcnn.layers.gather import GatherNodesOutgoing
from kgcnn.layers.mlp import GraphMLP
from kgcnn.layers.update import ResidualLayer
from kgcnn.initializers.initializers import GlorotOrthogonal, HeOrthogonal


class DimNetInteractionPPBlock(Layer):
    """DimNetPP Interaction Block as defined by `DimNetPP <https://arxiv.org/abs/2011.14115>`__ .

    Args:
        emb_size: Embedding size used for the messages
        int_emb_size (int): Embedding size used for interaction triplets
        basis_emb_size: Embedding size used inside the basis transformation
        num_before_skip: Number of residual layers in interaction block before skip connection
        num_after_skip: Number of residual layers in interaction block before skip connection
        use_bias (bool, optional): Use bias. Defaults to True.
        pooling_method (str): Pooling method information for layer. Default is 'sum'.
        activation (str): Activation function. Default is "swish".
        kernel_regularizer: Kernel regularization. Default is None.
        bias_regularizer: Bias regularization. Default is None.
        activity_regularizer: Activity regularization. Default is None.
        kernel_constraint: Kernel constrains. Default is None.
        bias_constraint: Bias constrains. Default is None.
        kernel_initializer: Initializer for kernels. Default is 'kgcnn>glorot_orthogonal'.
        bias_initializer: Initializer for bias. Default is 'zeros'.
    """

    def __init__(self, emb_size,
                 int_emb_size,
                 basis_emb_size,
                 num_before_skip,
                 num_after_skip,
                 use_bias=True,
                 pooling_method="sum",
                 activation='swish',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer="kgcnn>glorot_orthogonal",  # default is 'kgcnn>glorot_orthogonal'
                 bias_initializer='zeros',
                 **kwargs):
        super(DimNetInteractionPPBlock, self).__init__(**kwargs)
        self.use_bias = use_bias
        self.pooling_method = pooling_method
        self.emb_size = emb_size
        self.int_emb_size = int_emb_size
        self.basis_emb_size = basis_emb_size
        self.num_before_skip = num_before_skip
        self.num_after_skip = num_after_skip
        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
                       "bias_constraint": bias_constraint, "kernel_initializer": kernel_initializer,
                       "bias_initializer": bias_initializer}

        # Transformations of Bessel and spherical basis representations
        self.dense_rbf1 = Dense(basis_emb_size, use_bias=False, **kernel_args)
        self.dense_rbf2 = Dense(emb_size, use_bias=False, **kernel_args)
        self.dense_sbf1 = Dense(basis_emb_size, use_bias=False, **kernel_args)
        self.dense_sbf2 = Dense(int_emb_size, use_bias=False, **kernel_args)

        # Dense transformations of input messages
        self.dense_ji = Dense(emb_size, activation=activation, use_bias=True, **kernel_args)
        self.dense_kj = Dense(emb_size, activation=activation, use_bias=True, **kernel_args)

        # Embedding projections for interaction triplets
        self.down_projection = Dense(int_emb_size, activation=activation, use_bias=False, **kernel_args)
        self.up_projection = Dense(emb_size, activation=activation, use_bias=False, **kernel_args)

        # Residual layers before skip connection
        self.layers_before_skip = []
        for i in range(num_before_skip):
            self.layers_before_skip.append(
                ResidualLayer(emb_size, activation=activation, use_bias=True, **kernel_args))
        self.final_before_skip = Dense(emb_size, activation=activation, use_bias=True, **kernel_args)

        # Residual layers after skip connection
        self.layers_after_skip = []
        for i in range(num_after_skip):
            self.layers_after_skip.append(
                ResidualLayer(emb_size, activation=activation, use_bias=True, **kernel_args))

        self.lay_add1 = Add()
        self.lay_add2 = Add()
        self.lay_mult1 = Multiply()
        self.lay_mult2 = Multiply()

        self.lay_gather = GatherNodesOutgoing()  # Are edges here
        self.lay_pool = AggregateLocalEdges(pooling_method=pooling_method)

    def build(self, input_shape):
        """Build layer."""
        super(DimNetInteractionPPBlock, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [edges, rbf, sbf, angle_index]

                - edges (Tensor): Edge embeddings of shape ([M], F)
                - rbf (Tensor): Radial basis features of shape ([M], F)
                - sbf (Tensor): Spherical basis features of shape ([K], F)
                - angle_index (Tensor): Angle indices referring to two edges of shape (2, [K])

        Returns:
            Tensor: Updated edge embeddings.
        """
        x, rbf, sbf, id_expand = inputs

        # Initial transformation
        x_ji = self.dense_ji(x, **kwargs)
        x_kj = self.dense_kj(x, **kwargs)

        # Transform via Bessel basis
        rbf = self.dense_rbf1(rbf, **kwargs)
        rbf = self.dense_rbf2(rbf, **kwargs)
        x_kj = self.lay_mult1([x_kj, rbf], **kwargs)

        # Down-project embeddings and generate interaction triplet embeddings
        x_kj = self.down_projection(x_kj, **kwargs)
        x_kj = self.lay_gather([x_kj, id_expand], **kwargs)

        # Transform via 2D spherical basis
        sbf = self.dense_sbf1(sbf, **kwargs)
        sbf = self.dense_sbf2(sbf, **kwargs)
        x_kj = self.lay_mult2([x_kj, sbf], **kwargs)

        # Aggregate interactions and up-project embeddings
        x_kj = self.lay_pool([rbf, x_kj, id_expand], **kwargs)
        x_kj = self.up_projection(x_kj, **kwargs)

        # Transformations before skip connection
        x2 = self.lay_add1([x_ji, x_kj], **kwargs)
        for layer in self.layers_before_skip:
            x2 = layer(x2, **kwargs)
        x2 = self.final_before_skip(x2, **kwargs)

        # Skip connection
        x = self.lay_add2([x, x2],**kwargs)

        # Transformations after skip connection
        for layer in self.layers_after_skip:
            x = layer(x, **kwargs)

        return x

    def get_config(self):
        config = super(DimNetInteractionPPBlock, self).get_config()
        config.update({"use_bias": self.use_bias, "pooling_method": self.pooling_method, "emb_size": self.emb_size,
                       "int_emb_size": self.int_emb_size, "basis_emb_size": self.basis_emb_size,
                       "num_before_skip": self.num_before_skip, "num_after_skip": self.num_after_skip})
        conf_dense = self.dense_ji.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation"]:
            if x in conf_dense:
                config.update({x: conf_dense[x]})
        return config


class DimNetOutputBlock(Layer):
    """DimNetPP Output Block as defined by `DimNetPP <https://arxiv.org/abs/2011.14115>`__ .

    Args:
        emb_size (list): List of node embedding dimension.
        out_emb_size (list): List of edge embedding dimension.
        num_dense (list): Number of dense layer for MLP.
        num_targets (int): Number of output target dimension. Defaults to 12.
        use_bias (bool, optional): Use bias. Defaults to True.
        kernel_initializer: Initializer for kernels. Default is 'glorot_orthogonal' with fallback 'orthogonal'.
        output_kernel_initializer: Initializer for last kernel. Default is 'zeros'.
        bias_initializer: Initializer for bias. Default is 'zeros'.
        activation (str): Activation function. Default is 'kgcnn>swish'.
        kernel_regularizer: Kernel regularization. Default is None.
        bias_regularizer: Bias regularization. Default is None.
        activity_regularizer: Activity regularization. Default is None.
        kernel_constraint: Kernel constrains. Default is None.
        bias_constraint: Bias constrains. Default is None.
        pooling_method (str): Pooling method information for layer. Default is 'mean'.
    """

    def __init__(self, emb_size,
                 out_emb_size,
                 num_dense,
                 num_targets=12,
                 use_bias=True,
                 output_kernel_initializer="zeros", kernel_initializer='kgcnn>glorot_orthogonal',
                 bias_initializer='zeros',
                 activation='swish',
                 kernel_regularizer=None, bias_regularizer=None, activity_regularizer=None,
                 kernel_constraint=None, bias_constraint=None,
                 pooling_method="sum",
                 **kwargs):
        """Initialize layer."""
        super(DimNetOutputBlock, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.emb_size = emb_size
        self.out_emb_size = out_emb_size
        self.num_dense = num_dense
        self.num_targets = num_targets
        self.use_bias = use_bias
        kernel_args = {"kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
                       "kernel_constraint": kernel_constraint, "bias_initializer": bias_initializer,
                       "bias_regularizer": bias_regularizer, "bias_constraint": bias_constraint, }

        self.dense_rbf = Dense(emb_size, use_bias=False, kernel_initializer=kernel_initializer, **kernel_args)
        self.up_projection = Dense(out_emb_size, use_bias=False, kernel_initializer=kernel_initializer, **kernel_args)
        self.dense_mlp = GraphMLP([out_emb_size] * num_dense, activation=activation,
                                  kernel_initializer=kernel_initializer, use_bias=use_bias, **kernel_args)
        self.dimnet_mult = Multiply()
        self.pool = AggregateLocalEdges(pooling_method=self.pooling_method)
        self.dense_final = Dense(num_targets, use_bias=False, kernel_initializer=output_kernel_initializer,
                                 **kernel_args)

    def build(self, input_shape):
        """Build layer."""
        super(DimNetOutputBlock, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [nodes, edges, tensor_index, state]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - edges (Tensor): Edge or message embeddings of shape ([M], F)
                - rbf (Tensor): Edge distance basis of shape ([M], F)
                - tensor_index (Tensor): Edge indices referring to nodes of shape (2, [M])

        Returns:
            Tensor: Updated node embeddings of shape ([N], F_T).
        """
        # Calculate edge Update
        n_atoms, x, rbf, idnb_i = inputs
        g = self.dense_rbf(rbf, **kwargs)
        x = self.dimnet_mult([g, x], **kwargs)
        x = self.pool([n_atoms, x, idnb_i], **kwargs)
        x = self.up_projection(x, **kwargs)
        x = self.dense_mlp(x, **kwargs)
        x = self.dense_final(x, **kwargs)
        return x

    def get_config(self):
        config = super(DimNetOutputBlock, self).get_config()
        conf_mlp = self.dense_mlp.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation"]:
            if x in conf_mlp:
                config.update({x: conf_mlp[x][0]})
        conf_dense_output = self.dense_final.get_config()
        config.update({"output_kernel_initializer": conf_dense_output["kernel_initializer"]})
        config.update({"pooling_method": self.pooling_method, "use_bias": self.use_bias})
        config.update({"emb_size": self.emb_size, "out_emb_size": self.out_emb_size, "num_dense": self.num_dense,
                       "num_targets": self.num_targets})
        return config


class EmbeddingDimeBlock(Layer):
    """Custom Embedding Block of `DimNetPP <https://arxiv.org/abs/2011.14115>`__ .

    Naming of inputs here should match keras Embedding layer.

    Args:
        input_dim (int): Integer. Size of the vocabulary, i.e. maximum integer index + 1.
        output_dim (int): Integer. Dimension of the dense embedding.
        embeddings_initializer: Initializer for the embeddings matrix (see keras.initializers).
        embeddings_regularizer: Regularizer function applied to the embeddings matrix (see keras.regularizers).
        embeddings_constraint: Constraint function applied to the embeddings matrix (see keras.constraints).

    """
    def __init__(self,
                 input_dim,  # Vocabulary
                 output_dim,  # Embedding size
                 embeddings_initializer='uniform',
                 embeddings_regularizer=None,
                 embeddings_constraint=None,
                 **kwargs):
        super(EmbeddingDimeBlock, self).__init__(**kwargs)
        self._supports_ragged_inputs = True
        self.output_dim = output_dim
        self.input_dim = input_dim
        self.embeddings_initializer = ks.initializers.get(embeddings_initializer)
        self.embeddings_regularizer = ks.regularizers.get(embeddings_regularizer)
        self.embeddings_constraint = ks.constraints.get(embeddings_constraint)

        # Original implementation used initializer:
        # embeddings_initializer = {'class_name': 'RandomUniform', 'config': {'minval': -1.7320508075688772,
        # 'maxval': 1.7320508075688772, 'seed': None}}
        self.embeddings = self.add_weight(name="embeddings", shape=(self.input_dim + 1, self.output_dim),
                                          dtype=self.dtype, initializer=self.embeddings_initializer,
                                          regularizer=self.embeddings_regularizer,
                                          constraint=self.embeddings_constraint,
                                          trainable=True)

    def call(self, inputs, **kwargs):
        """Embedding of inputs. Forward pass."""
        out = ops.take(self.embeddings, inputs, axis=0)
        return out

    def get_config(self):
        config = super(EmbeddingDimeBlock, self).get_config()
        config.update({"input_dim": self.input_dim, "output_dim": self.output_dim,
                       "embeddings_initializer": ks.initializers.serialize(self.embeddings_initializer),
                       "embeddings_regularizer": ks.regularizers.serialize(self.embeddings_regularizer),
                       "embeddings_constraint": ks.constraints.serialize(self.embeddings_constraint)
                       })
        return config



from keras.layers import Add
from kgcnn.layers.geom import NodePosition, EdgeDirectionNormalized, NodeDistanceEuclidean, CosCutOffEnvelope, \
    BesselBasisLayer, ShiftPeriodicLattice
from kgcnn.layers.mlp import MLP, GraphMLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.norm import GraphLayerNormalization, GraphBatchNormalization
from kgcnn.layers.pooling import PoolingNodes
from ._layers import EquivariantInitialize, PAiNNconv, PAiNNUpdate


def model_disjoint(
        inputs: list,
        use_node_embedding: bool,
        input_node_embedding: dict,
        equiv_initialize_kwargs: dict,
        bessel_basis: dict,
        depth: int,
        pooling_args: dict,
        conv_args: dict,
        update_args: dict,
        equiv_normalization: bool,
        node_normalization: bool,
        output_embedding: str,
        output_mlp: dict,
):
    z, x, edi, batch_id_node, batch_id_edge, count_nodes, count_edges, v = inputs

    if v is None:
        v = EquivariantInitialize(**equiv_initialize_kwargs)(z)

    # Optional Embedding.
    if use_node_embedding:
        z = Embedding(**input_node_embedding)(z)

    pos1, pos2 = NodePosition()([x, edi])
    rij = EdgeDirectionNormalized()([pos1, pos2])
    d = NodeDistanceEuclidean()([pos1, pos2])
    env = CosCutOffEnvelope(conv_args["cutoff"])(d)
    rbf = BesselBasisLayer(**bessel_basis)(d)

    for i in range(depth):
        # Message
        ds, dv = PAiNNconv(**conv_args)([z, v, rbf, env, rij, edi])
        z = Add()([z, ds])
        v = Add()([v, dv])
        # Update
        ds, dv = PAiNNUpdate(**update_args)([z, v])
        z = Add()([z, ds])
        v = Add()([v, dv])

        if equiv_normalization:
            v = GraphLayerNormalization(axis=2)([v, batch_id_edge, count_edges])
        if node_normalization:
            z = GraphBatchNormalization(axis=-1)([z, batch_id_node, count_nodes])

    n = z
    # Output embedding choice
    if output_embedding == "graph":
        out = PoolingNodes(**pooling_args)([count_nodes, n, batch_id_node])
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `PAiNN`")

    return out


def model_disjoint_crystal(
        inputs: list,
        use_node_embedding: bool,
        input_node_embedding: dict,
        equiv_initialize_kwargs: dict,
        bessel_basis: dict,
        depth: int,
        pooling_args: dict,
        conv_args: dict,
        update_args: dict,
        equiv_normalization: bool,
        node_normalization: bool,
        output_embedding: str,
        output_mlp: dict,
):
    z, x, edi, edge_image, lattice, batch_id_node, batch_id_edge, count_nodes, count_edges, v = inputs

    if v is None:
        v = EquivariantInitialize(**equiv_initialize_kwargs)(z)

    # Optional Embedding.
    if use_node_embedding:
        z = Embedding(**input_node_embedding)(z)

    pos1, pos2 = NodePosition()([x, edi])
    pos2 = ShiftPeriodicLattice()([pos2, edge_image, lattice, batch_id_edge])
    rij = EdgeDirectionNormalized()([pos1, pos2])
    d = NodeDistanceEuclidean()([pos1, pos2])
    env = CosCutOffEnvelope(conv_args["cutoff"])(d)
    rbf = BesselBasisLayer(**bessel_basis)(d)

    for i in range(depth):
        # Message
        ds, dv = PAiNNconv(**conv_args)([z, v, rbf, env, rij, edi])
        z = Add()([z, ds])
        v = Add()([v, dv])
        # Update
        ds, dv = PAiNNUpdate(**update_args)([z, v])
        z = Add()([z, ds])
        v = Add()([v, dv])

        if equiv_normalization:
            v = GraphLayerNormalization(axis=2)([v, batch_id_edge, count_edges])
        if node_normalization:
            z = GraphBatchNormalization(axis=-1)([z, batch_id_node, count_nodes])

    n = z
    # Output embedding choice
    if output_embedding == "graph":
        out = PoolingNodes(**pooling_args)([count_nodes, n, batch_id_node])
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `PAiNN`")

    return out


from kgcnn.layers.modules import Embedding
from kgcnn.layers.geom import NodePosition, NodeDistanceEuclidean, GaussBasisLayer, ShiftPeriodicLattice
from kgcnn.layers.mlp import MLP, GraphMLP
from keras.layers import Dense, Dropout, Concatenate, Flatten, Add
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.set2set import PoolingSet2SetEncoder
from ._layers import MEGnetBlock


PoolingGlobalEdges = PoolingNodes


def model_disjoint(
        inputs,
        use_node_embedding,
        use_graph_embedding,
        input_node_embedding: dict = None,
        input_graph_embedding: dict = None,
        expand_distance: bool = None,
        make_distance: bool = None,
        gauss_args: dict = None,
        meg_block_args: dict = None,
        set2set_args: dict = None,
        node_ff_args: dict = None,
        edge_ff_args: dict = None,
        state_ff_args: dict = None,
        use_set2set: bool = None,
        nblocks: int = None,
        has_ff: bool = None,
        dropout: float = None,
        output_embedding: str = None,
        output_mlp: dict = None,
):
    # Make input
    vp, x, edi, up, batch_id_node, batch_id_edge, count_nodes, count_edges = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        vp = Embedding(**input_node_embedding)(vp)
    if use_graph_embedding:
        up = Embedding(**input_graph_embedding)(up)

    # Edge distance as Gauss-Basis
    if make_distance:
        pos1, pos2 = NodePosition()([x, edi])
        ep = NodeDistanceEuclidean()([pos1, pos2])
    else:
        ep = x

    if expand_distance:
        ep = GaussBasisLayer(**gauss_args)(ep)

    # Model
    vp = GraphMLP(**node_ff_args)([vp, batch_id_node, count_nodes])
    ep = GraphMLP(**edge_ff_args)([ep, batch_id_edge, count_edges])
    up = MLP(**state_ff_args)(up)
    vp2 = vp
    ep2 = ep
    up2 = up
    for i in range(0, nblocks):
        if has_ff and i > 0:
            vp2 = GraphMLP(**node_ff_args)([vp, batch_id_node, count_nodes])
            ep2 = GraphMLP(**edge_ff_args)([ep, batch_id_edge, count_edges])
            up2 = MLP(**state_ff_args)(up)

        # MEGnetBlock
        vp2, ep2, up2 = MEGnetBlock(**meg_block_args)(
            [vp2, ep2, edi, up2, batch_id_node, batch_id_edge, count_nodes, count_edges])

        # skip connection
        if dropout is not None:
            vp2 = Dropout(dropout, name='dropout_atom_%d' % i)(vp2)
            ep2 = Dropout(dropout, name='dropout_bond_%d' % i)(ep2)
            up2 = Dropout(dropout, name='dropout_state_%d' % i)(up2)

        vp = Add()([vp2, vp])
        ep = Add()([ep2, ep])
        up = Add()([up2, up])

    if use_set2set:
        vp = Dense(set2set_args["channels"], activation='linear')(vp)  # to match units
        ep = Dense(set2set_args["channels"], activation='linear')(ep)  # to match units
        vp = PoolingSet2SetEncoder(**set2set_args)([count_nodes, vp, batch_id_node])
        ep = PoolingSet2SetEncoder(**set2set_args)([count_edges, ep, batch_id_edge])
    else:
        vp = PoolingNodes()([count_nodes, vp, batch_id_node])
        ep = PoolingGlobalEdges()([count_edges, ep, batch_id_edge])

    ep = Flatten()(ep)
    vp = Flatten()(vp)
    final_vec = Concatenate(axis=-1)([vp, ep, up])

    if dropout is not None:
        final_vec = Dropout(dropout, name='dropout_final')(final_vec)

    # Only graph embedding for Megnet
    if output_embedding != "graph":
        raise ValueError("Unsupported output embedding for mode `Megnet`.")

    main_output = MLP(**output_mlp)(final_vec)
    return main_output


def model_disjoint_crystal(
        inputs,
        use_node_embedding,
        use_graph_embedding,
        input_node_embedding: dict = None,
        input_graph_embedding: dict = None,
        expand_distance: bool = None,
        make_distance: bool = None,
        gauss_args: dict = None,
        meg_block_args: dict = None,
        set2set_args: dict = None,
        node_ff_args: dict = None,
        edge_ff_args: dict = None,
        state_ff_args: dict = None,
        use_set2set: bool = None,
        nblocks: int = None,
        has_ff: bool = None,
        dropout: float = None,
        output_embedding: str = None,
        output_mlp: dict = None,
):
    vp, x, edi, up, edge_image, lattice, batch_id_node, batch_id_edge, count_nodes, count_edges = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        vp = Embedding(**input_node_embedding)(vp)
    if use_graph_embedding:
        up = Embedding(**input_graph_embedding)(up)

    # Edge distance as Gauss-Basis
    if make_distance:
        pos1, pos2 = NodePosition()([x, edi])
        pos2 = ShiftPeriodicLattice()([pos2, edge_image, lattice, batch_id_edge])
        ep = NodeDistanceEuclidean()([pos1, pos2])
    else:
        ep = x

    if expand_distance:
        ep = GaussBasisLayer(**gauss_args)(ep)

    # Model
    vp = GraphMLP(**node_ff_args)([vp, batch_id_edge, count_edges])
    ep = GraphMLP(**edge_ff_args)([ep, batch_id_edge, count_edges])
    up = MLP(**state_ff_args)(up)
    vp2 = vp
    ep2 = ep
    up2 = up
    for i in range(0, nblocks):
        if has_ff and i > 0:
            vp2 = GraphMLP(**node_ff_args)([vp, batch_id_node, count_nodes])
            ep2 = GraphMLP(**edge_ff_args)([ep, batch_id_edge, count_edges])
            up2 = MLP(**state_ff_args)(up)

        # MEGnetBlock
        vp2, ep2, up2 = MEGnetBlock(**meg_block_args)(
            [vp2, ep2, edi, up2, batch_id_node, batch_id_edge, count_nodes, count_edges])

        # skip connection
        if dropout is not None:
            vp2 = Dropout(dropout, name='dropout_atom_%d' % i)(vp2)
            ep2 = Dropout(dropout, name='dropout_bond_%d' % i)(ep2)
            up2 = Dropout(dropout, name='dropout_state_%d' % i)(up2)

        vp = Add()([vp2, vp])
        ep = Add()([ep2, ep])
        up = Add()([up2, up])

    if use_set2set:
        vp = Dense(set2set_args["channels"], activation='linear')(vp)  # to match units
        ep = Dense(set2set_args["channels"], activation='linear')(ep)  # to match units
        vp = PoolingSet2SetEncoder(**set2set_args)([count_nodes, vp, batch_id_node])
        ep = PoolingSet2SetEncoder(**set2set_args)([count_edges, ep, batch_id_edge])
    else:
        vp = PoolingNodes()([count_nodes, vp, batch_id_node])
        ep = PoolingGlobalEdges()([count_edges, ep, batch_id_edge])

    ep = Flatten()(ep)
    vp = Flatten()(vp)
    final_vec = Concatenate(axis=-1)([vp, ep, up])

    if dropout is not None:
        final_vec = Dropout(dropout, name='dropout_final')(final_vec)

    # Only graph embedding for Megnet
    if output_embedding != "graph":
        raise ValueError("Unsupported output embedding for mode `Megnet`.")

    main_output = MLP(**output_mlp)(final_vec)

    return main_output



import keras as ks
import itertools
from keras.layers import Concatenate
from kgcnn.layers.mlp import MLP


def merge_models(model_list: list,
                 merge_type: str = "concat",
                 output_mlp: dict = None):
    r"""Merge a list of models by combining their output.

    Args:
        model_list (list): List of graph models.
        merge_type (str): How to merge the output.
        output_mlp (dict): Kwargs of the final MLP after the models' output.

    Returns:
        :obj:`ks.models.Model`
    """
    if output_mlp:
        if isinstance(output_mlp, dict):
            output_mlp = MLP(**output_mlp)

    combined_inputs = []
    for m in model_list:
        new_inputs_per_model = []
        for i, input_layer in enumerate(m.inputs):
            new_input_layer = ks.Input(type_spec=input_layer.type_spec, name=input_layer.name)
            new_inputs_per_model.append(new_input_layer)
        combined_inputs.append(new_inputs_per_model)

    new_outputs = []
    for x, m in zip(combined_inputs, model_list):
        new_outputs.append(m(x))

    if merge_type in ["concat", "concatenate"]:
        output = Concatenate(axis=-1)(new_outputs)
    else:
        raise NotImplementedError("Unknown merge type '%s' for models" % merge_type)

    if output_mlp:
        output = output_mlp(output)

    flatten_inputs = list(itertools.chain(*combined_inputs))
    return ks.models.Model(inputs=flatten_inputs, outputs=output)



import keras as ks
from kgcnn.layers.scale import get as get_scaler
from ._model import model_disjoint
from kgcnn.layers.modules import Input
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from kgcnn.models.utils import update_model_kwargs
from keras.backend import backend as backend_to_use

# To be updated if model is changed in a significant way.
__model_version__ = "2023-12-09"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]
if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'MXMNet' is not supported." % backend_to_use())

# Implementation of MXMNet in `tf.keras` from paper:
# Molecular Mechanics-Driven Graph Neural Network with Multiplex Graph for Molecular Structures
# by Shuo Zhang, Yang Liu, Lei Xie (2020)
# https://arxiv.org/abs/2011.07457
# https://github.com/zetayue/MXMNet


model_default = {
    "name": "MXMNet",
    "inputs": [
        {"shape": (None, ), "name": "node_number", "dtype": "float32"},
        {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32"},
        {"shape": (None, 64), "name": "edge_attributes", "dtype": "float32"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (None, 2), "name": "range_indices", "dtype": "int64"},
        {"shape": [None, 2], "name": "angle_indices_1", "dtype": "int64"},
        {"shape": [None, 2], "name": "angle_indices_2", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"},
        {"shape": (), "name": "total_ranges", "dtype": "int64"},
        {"shape": (), "name": "total_angles_1", "dtype": "int64"},
        {"shape": (), "name": "total_angles_2", "dtype": "int64"}
    ],
    "input_tensor_type": "padded",
    "input_embedding": None,  # deprecated
    "cast_disjoint_kwargs": {},
    "input_node_embedding": {
        "input_dim": 95, "output_dim": 32,
        "embeddings_initializer": {
            "class_name": "RandomUniform",
            "config": {"minval": -1.7320508075688772, "maxval": 1.7320508075688772}}
    },
    "input_edge_embedding": {"input_dim": 32, "output_dim": 32},
    "bessel_basis_local": {"num_radial": 16, "cutoff": 5.0, "envelope_exponent": 5},
    "bessel_basis_global": {"num_radial": 16, "cutoff": 5.0, "envelope_exponent": 5},  # Should match range_indices
    "spherical_basis_local": {"num_spherical": 7, "num_radial": 6, "cutoff": 5.0, "envelope_exponent": 5},
    "mlp_rbf_kwargs": {"units": 32, "activation": "swish"},
    "mlp_sbf_kwargs": {"units": 32, "activation": "swish"},
    "global_mp_kwargs": {"units": 32},
    "local_mp_kwargs": {"units": 32, "output_units": 1, "output_kernel_initializer": "zeros"},
    "use_edge_attributes": False,
    "depth": 3,
    "verbose": 10,
    "node_pooling_args": {"pooling_method": "sum"},
    "output_embedding": "graph",
    "use_output_mlp": True,
    "output_mlp": {"use_bias": [True], "units": [1],
                   "activation": ["linear"]},
    "output_tensor_type": "padded",
    "output_scaling": None,
    "output_to_tensor": None  # deprecated
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(inputs: list = None,
               input_tensor_type: str = None,
               cast_disjoint_kwargs: dict = None,
               input_embedding: dict = None,
               input_node_embedding: dict = None,
               input_edge_embedding: dict = None,
               depth: int = None,
               name: str = None,
               bessel_basis_local: dict = None,
               bessel_basis_global: dict = None,
               spherical_basis_local: dict = None,
               use_edge_attributes: bool = None,
               mlp_rbf_kwargs: dict = None,
               mlp_sbf_kwargs: dict = None,
               global_mp_kwargs: dict = None,
               local_mp_kwargs: dict = None,
               verbose: int = None,
               output_embedding: str = None,
               use_output_mlp: bool = None,
               node_pooling_args: dict = None,
               output_to_tensor: bool = None,
               output_mlp: dict = None,
               output_scaling: dict = None,
               output_tensor_type: str = None,
               ):
    r"""Make `MXMNet <https://arxiv.org/abs/2011.07457>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.MXMNet.model_default` .

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are
    :obj:`[nodes, coordinates, edge_attributes, edge_indices, range_indices, angle_indices_1, angle_indices_2, ...]`
    with '...' indicating mask or ID tensors following the template below.
    Note that you must supply angle indices as index pairs that refer to two edges or two range connections.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layer.
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        input_node_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        input_edge_embedding (dict): Dictionary of embedding arguments for nodes unpacked in :obj:`Embedding` layers.
        depth (int): Number of graph embedding units or depth of the network.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        bessel_basis_local: Dictionary of layer arguments unpacked in local `:obj:BesselBasisLayer` layer.
        bessel_basis_global: Dictionary of layer arguments unpacked in global `:obj:BesselBasisLayer` layer.
        spherical_basis_local: Dictionary of layer arguments unpacked in `:obj:SphericalBasisLayer` layer.
        use_edge_attributes: Whether to add edge attributes. Default is False.
        mlp_rbf_kwargs: Dictionary of layer arguments unpacked in `:obj:MLP` layer for RBF feed-forward.
        mlp_sbf_kwargs: Dictionary of layer arguments unpacked in `:obj:MLP` layer for SBF feed-forward.
        global_mp_kwargs: Dictionary of layer arguments unpacked in `:obj:MXMGlobalMP` layer.
        local_mp_kwargs: Dictionary of layer arguments unpacked in `:obj:MXMLocalMP` layer.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        output_to_tensor (bool): Whether to cast model output to :obj:`tf.Tensor` .
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers. Default is None.
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    dj = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        mask_assignment=[0, 0, 1, 1, 2, 3, 4],
        index_assignment=[None, None, None, 0, 0, 3, 3]
    )

    n, x, ed, edi, rgi, adi1, adi2 = dj[:7]
    batch_id_node, batch_id_edge, batch_id_ranges, batch_id_angles_1, batch_id_angles_2 = dj[7:12]
    node_id, edge_id, range_id, angle_id1, angle_id2 = dj[12:17]
    count_nodes, count_edges, count_ranges, count_angles1, count_angles2 = dj[17:]

    out = model_disjoint(
        dj,
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        use_edge_embedding=("int" in inputs[2]['dtype']) if input_edge_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        input_edge_embedding=input_edge_embedding,
        bessel_basis_local=bessel_basis_local,
        spherical_basis_local=spherical_basis_local,
        bessel_basis_global=bessel_basis_global,
        use_edge_attributes=use_edge_attributes,
        mlp_rbf_kwargs=mlp_rbf_kwargs,
        mlp_sbf_kwargs=mlp_sbf_kwargs,
        depth=depth,
        global_mp_kwargs=global_mp_kwargs,
        local_mp_kwargs=local_mp_kwargs,
        node_pooling_args=node_pooling_args,
        output_embedding=output_embedding,
        use_output_mlp=use_output_mlp,
        output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out = scaler([out, n, batch_id_node])
        else:
            out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



import keras as ks
from kgcnn.layers.scale import get as get_scaler
from ._model import model_disjoint
from kgcnn.layers.modules import Input
from kgcnn.models.casting import (template_cast_output, template_cast_list_input,
                                  template_cast_list_input_docs, template_cast_output_docs)
from kgcnn.models.utils import update_model_kwargs
from keras.backend import backend as backend_to_use

# To be updated if model is changed in a significant way.
__model_version__ = "2023-12-04"

# Supported backends
__kgcnn_model_backend_supported__ = ["tensorflow", "torch", "jax"]
if backend_to_use() not in __kgcnn_model_backend_supported__:
    raise NotImplementedError("Backend '%s' for model 'EGNN' is not supported." % backend_to_use())

# Implementation of EGNN in `keras` from paper:
# E(n) Equivariant Graph Neural Networks
# by Victor Garcia Satorras, Emiel Hoogeboom, Max Welling (2021)
# https://arxiv.org/abs/2102.09844


model_default = {
    "name": "EGNN",
    "inputs": [
        {"shape": (None,), "name": "node_number", "dtype": "int64"},
        {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32"},
        {"shape": (None, 10), "name": "edge_attributes", "dtype": "float32"},
        {"shape": (None, 2), "name": "edge_indices", "dtype": "int64"},
        {"shape": (), "name": "total_nodes", "dtype": "int64"},
        {"shape": (), "name": "total_edges", "dtype": "int64"},
    ],
    "input_tensor_type": "padded",
    "cast_disjoint_kwargs": {},
    "input_embedding": None,
    "input_node_embedding": {"input_dim": 95, "output_dim": 64},
    "input_edge_embedding": {"input_dim": 95, "output_dim": 64},
    "depth": 4,
    "node_mlp_initialize": None,
    "euclidean_norm_kwargs": {"keepdims": True, "axis": -1},
    "use_edge_attributes": True,
    "edge_mlp_kwargs": {"units": [64, 64], "activation": ["swish", "linear"]},
    "edge_attention_kwargs": None,  # {"units: 1", "activation": "sigmoid"}
    "use_normalized_difference": False,
    "expand_distance_kwargs": None,
    "coord_mlp_kwargs": {"units": [64, 1], "activation": ["swish", "linear"]},  # option: "tanh" at the end.
    "pooling_coord_kwargs": {"pooling_method": "mean"},
    "pooling_edge_kwargs": {"pooling_method": "sum"},
    "node_normalize_kwargs": None,
    "use_node_attributes": False,
    "node_mlp_kwargs": {"units": [64, 64], "activation": ["swish", "linear"]},
    "use_skip": True,
    "verbose": 10,
    "node_decoder_kwargs": None,
    "node_pooling_kwargs": {"pooling_method": "sum"},
    "output_embedding": "graph",
    "output_to_tensor": None,  # deprecated
    "output_tensor_type": "padded",
    "output_mlp": {"use_bias": [True, True], "units": [64, 1],
                   "activation": ["swish", "linear"]},
    "output_scaling": None,
}


@update_model_kwargs(model_default, update_recursive=0, deprecated=["input_embedding", "output_to_tensor"])
def make_model(name: str = None,
               inputs: list = None,
               input_tensor_type: str = None,
               cast_disjoint_kwargs: dict = None,
               input_embedding: dict = None,
               input_node_embedding: dict = None,
               input_edge_embedding: dict = None,
               depth: int = None,
               euclidean_norm_kwargs: dict = None,
               node_mlp_initialize: dict = None,
               use_edge_attributes: bool = None,
               edge_mlp_kwargs: dict = None,
               edge_attention_kwargs: dict = None,
               use_normalized_difference: bool = None,
               expand_distance_kwargs: dict = None,
               coord_mlp_kwargs: dict = None,
               pooling_coord_kwargs: dict = None,
               pooling_edge_kwargs: dict = None,
               node_normalize_kwargs: dict = None,
               use_node_attributes: bool = None,
               node_mlp_kwargs: dict = None,
               use_skip: bool = None,
               verbose: int = None,
               node_decoder_kwargs: dict = None,
               node_pooling_kwargs: dict = None,
               output_embedding: str = None,
               output_to_tensor: bool = None,
               output_mlp: dict = None,
               output_tensor_type: str = None,
               output_scaling: dict = None
               ):
    r"""Make `EGNN <https://arxiv.org/abs/2102.09844>`__ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.EGNN.model_default`.

    **Model inputs**:
    Model uses the list template of inputs and standard output template.
    The supported inputs are  :obj:`[nodes, node_coordinates, edge_attributes, edge_indices, ...]`
    with '...' indicating mask or ID tensors following the template below.

    %s

    **Model outputs**:
    The standard output template:

    %s

    Args:
        name (str): Name of the model. Default is "EGNN".
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        cast_disjoint_kwargs (dict): Dictionary of arguments for casting layers if used.
        input_tensor_type (str): Input type of graph tensor. Default is "padded".
        input_embedding (dict): Deprecated in favour of input_node_embedding etc.
        input_node_embedding (dict): Dictionary of arguments for nodes unpacked in :obj:`Embedding` layers.
        input_edge_embedding (dict): Dictionary of arguments for edge unpacked in :obj:`Embedding` layers.
        depth (int): Number of graph embedding units or depth of the network.
        euclidean_norm_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`EuclideanNorm`.
        node_mlp_initialize (dict): Dictionary of layer arguments unpacked in :obj:`GraphMLP` layer for start embedding.
        use_edge_attributes (bool): Whether to use edge attributes including for example further edge information.
        edge_mlp_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`GraphMLP` layer.
        edge_attention_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`GraphMLP` layer.
        use_normalized_difference (bool): Whether to use a normalized difference vector for nodes.
        expand_distance_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`PositionEncodingBasisLayer`.
        coord_mlp_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`GraphMLP` layer.
        pooling_coord_kwargs (dict):
        pooling_edge_kwargs (dict):
        node_normalize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`GraphLayerNormalization` layer.
        use_node_attributes (bool): Whether to add node attributes before node MLP.
        node_mlp_kwargs (dict):
        use_skip (bool):
        verbose (int): Level of verbosity.
        node_decoder_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`MLP` layer after graph network.
        node_pooling_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): Deprecated in favour of `output_tensor_type` .
        output_tensor_type (str): Output type of graph tensors such as nodes or edges. Default is "padded".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.
        output_scaling (dict): Dictionary of layer arguments unpacked in scaling layers.

    Returns:
        :obj:`keras.models.Model`
    """
    # Make input
    model_inputs = [Input(**x) for x in inputs]

    dj = template_cast_list_input(
        model_inputs,
        input_tensor_type=input_tensor_type,
        cast_disjoint_kwargs=cast_disjoint_kwargs,
        index_assignment=[None, None, None, 0],
        mask_assignment=[0, 0, 1, 1]
    )

    n, x, ed, disjoint_indices, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = dj

    out = model_disjoint(
        [n, x, ed, disjoint_indices, batch_id_node, batch_id_edge, count_nodes, count_edges],
        use_node_embedding=("int" in inputs[0]['dtype']) if input_node_embedding is not None else False,
        use_edge_embedding=("int" in inputs[2]['dtype']) if input_edge_embedding is not None else False,
        input_node_embedding=input_node_embedding,
        input_edge_embedding=input_edge_embedding,
        depth=depth,
        euclidean_norm_kwargs=euclidean_norm_kwargs,
        node_mlp_initialize=node_mlp_initialize,
        use_edge_attributes=use_edge_attributes,
        edge_mlp_kwargs=edge_mlp_kwargs,
        edge_attention_kwargs=edge_attention_kwargs,
        use_normalized_difference=use_normalized_difference,
        expand_distance_kwargs=expand_distance_kwargs,
        coord_mlp_kwargs=coord_mlp_kwargs,
        pooling_coord_kwargs=pooling_coord_kwargs,
        pooling_edge_kwargs=pooling_edge_kwargs,
        node_normalize_kwargs=node_normalize_kwargs,
        use_node_attributes=use_node_attributes,
        node_mlp_kwargs=node_mlp_kwargs,
        use_skip=use_skip,
        node_decoder_kwargs=node_decoder_kwargs,
        node_pooling_kwargs=node_pooling_kwargs,
        output_embedding=output_embedding,
        output_mlp=output_mlp
    )

    if output_scaling is not None:
        scaler = get_scaler(output_scaling["name"])(**output_scaling)
        if scaler.extensive:
            # Node information must be numbers, or we need an additional input.
            out = scaler([out, n, batch_id_node])
        else:
            out = scaler(out)

    # Output embedding choice
    out = template_cast_output(
        [out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges],
        output_embedding=output_embedding, output_tensor_type=output_tensor_type,
        input_tensor_type=input_tensor_type, cast_disjoint_kwargs=cast_disjoint_kwargs,
    )

    model = ks.models.Model(inputs=model_inputs, outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__

    if output_scaling is not None:
        def set_scale(*args, **kwargs):
            scaler.set_scale(*args, **kwargs)

        setattr(model, "set_scale", set_scale)

    return model


make_model.__doc__ = make_model.__doc__ % (template_cast_list_input_docs, template_cast_output_docs)



import keras as ks
from ._layers import TrafoEdgeNetMessages
from kgcnn.layers.aggr import AggregateLocalEdges
from kgcnn.layers.gather import GatherNodesOutgoing, GatherNodesIngoing
from kgcnn.layers.geom import NodePosition, NodeDistanceEuclidean, GaussBasisLayer, ShiftPeriodicLattice
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.message import MatMulMessages
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.update import GRUUpdate
from kgcnn.layers.set2set import PoolingSet2SetEncoder


def model_disjoint(inputs,
                   use_node_embedding: bool = None,
                   use_edge_embedding: bool = None,
                   input_node_embedding: dict = None,
                   input_edge_embedding: dict = None,
                   geometric_edge: bool = None,
                   make_distance: bool = None,
                   expand_distance: bool = None,
                   gauss_args: dict = None,
                   set2set_args: dict = None,
                   pooling_args: dict = None,
                   edge_mlp: dict = None,
                   use_set2set: bool = None,
                   node_dim: int = None,
                   depth: int = None,
                   output_embedding: str = None,
                   output_mlp: dict = None):
    n0, ed, disjoint_indices, batch_id_node, batch_id_edge, count_nodes, count_edges = inputs

    # embedding, if no feature dimension
    if use_node_embedding:
        n0 = Embedding(**input_node_embedding)(n0)

    if not geometric_edge:
        if use_edge_embedding:
            ed = Embedding(**input_edge_embedding)(ed)

    if make_distance:
        pos1, pos2 = NodePosition()([ed, disjoint_indices])
        ed = NodeDistanceEuclidean()([pos1, pos2])

    if expand_distance:
        ed = GaussBasisLayer(**gauss_args)(ed)

    # Make hidden dimension
    n = ks.layers.Dense(node_dim, activation="linear")(n0)

    # Make edge networks.
    edge_net_in = GraphMLP(**edge_mlp)([ed, batch_id_edge, count_edges])
    edge_net_in = TrafoEdgeNetMessages(target_shape=(node_dim, node_dim))(edge_net_in)
    edge_net_out = GraphMLP(**edge_mlp)([ed, batch_id_edge, count_edges])
    edge_net_out = TrafoEdgeNetMessages(target_shape=(node_dim, node_dim))(edge_net_out)

    # Gru for node updates
    gru = GRUUpdate(node_dim)

    for i in range(0, depth):
        n_in = GatherNodesOutgoing()([n, disjoint_indices])
        n_out = GatherNodesIngoing()([n, disjoint_indices])
        m_in = MatMulMessages()([edge_net_in, n_in])
        m_out = MatMulMessages()([edge_net_out, n_out])
        eu = ks.layers.Concatenate(axis=-1)([m_in, m_out])
        eu = AggregateLocalEdges(**pooling_args)([n, eu, disjoint_indices])  # Summing for each node connections
        n = gru([n, eu])

    n = ks.layers.Concatenate(axis=-1)([n0, n])

    # Output embedding choice
    if output_embedding == 'graph':
        if use_set2set:
            # output
            n = ks.layers.Dense(units=set2set_args['channels'], activation="linear")(n)
            out = PoolingSet2SetEncoder(**set2set_args)([count_nodes, n, batch_id_node])
        else:
            out = PoolingNodes(**pooling_args)([count_nodes, n, batch_id_node])
        out = ks.layers.Flatten()(out)  # Flatten() required for to Set2Set output.
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `NMPN` .")

    return out


def model_disjoint_crystal(inputs,
                           use_node_embedding: bool = None,
                           use_edge_embedding: bool = None,
                           input_node_embedding: dict = None,
                           input_edge_embedding: dict = None,
                           geometric_edge: bool = None,
                           make_distance: bool = None,
                           expand_distance: bool = None,
                           gauss_args: dict = None,
                           set2set_args: dict = None,
                           pooling_args: dict = None,
                           edge_mlp: dict = None,
                           use_set2set: bool = None,
                           node_dim: int = None,
                           depth: int = None,
                           output_embedding: str = None,
                           output_mlp: dict = None):
    n0, ed, disjoint_indices, edge_image, lattice, batch_id_node, batch_id_edge, count_nodes, count_edges = inputs

    # embedding, if no feature dimension
    if use_node_embedding:
        n0 = Embedding(**input_node_embedding)(n0)

    if not geometric_edge:
        if use_edge_embedding:
            ed = Embedding(**input_edge_embedding)(ed)

    # If coordinates are in place of edges
    if make_distance:
        x = ed
        pos1, pos2 = NodePosition()([x, disjoint_indices])
        pos2 = ShiftPeriodicLattice()([pos2, edge_image, lattice, batch_id_edge])
        ed = NodeDistanceEuclidean()([pos1, pos2])

    if expand_distance:
        ed = GaussBasisLayer(**gauss_args)(ed)

    # Make hidden dimension
    n = ks.layers.Dense(node_dim, activation="linear")(n0)

    # Make edge networks.
    edge_net_in = GraphMLP(**edge_mlp)([ed, batch_id_edge, count_edges])
    edge_net_in = TrafoEdgeNetMessages(target_shape=(node_dim, node_dim))(edge_net_in)
    edge_net_out = GraphMLP(**edge_mlp)([ed, batch_id_edge, count_edges])
    edge_net_out = TrafoEdgeNetMessages(target_shape=(node_dim, node_dim))(edge_net_out)

    # Gru for node updates
    gru = GRUUpdate(node_dim)

    for i in range(0, depth):
        n_in = GatherNodesOutgoing()([n, disjoint_indices])
        n_out = GatherNodesIngoing()([n, disjoint_indices])
        m_in = MatMulMessages()([edge_net_in, n_in])
        m_out = MatMulMessages()([edge_net_out, n_out])
        eu = ks.layers.Concatenate(axis=-1)([m_in, m_out])
        eu = AggregateLocalEdges(**pooling_args)([n, eu, disjoint_indices])  # Summing for each node connections
        n = gru([n, eu])

    n = ks.layers.Concatenate(axis=-1)([n0, n])

    # Output embedding choice
    if output_embedding == 'graph':
        if use_set2set:
            # output
            n = ks.layers.Dense(units=set2set_args['channels'], activation="linear")(n)
            out = PoolingSet2SetEncoder(**set2set_args)([count_nodes, n, batch_id_node])
        else:
            out = PoolingNodes(**pooling_args)([count_nodes, n, batch_id_node])
        out = ks.layers.Flatten()(out)  # Flatten() required for to Set2Set output.
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `NMPN` .")

    return out



import numpy as np
from typing import Union
from pymatgen.optimization.neighbors import find_points_in_spheres


def range_neighbour_lattice(coordinates: np.ndarray, lattice: np.ndarray,
                            max_distance: Union[float, None] = 4.0,
                            max_neighbours: Union[int, None] = None,
                            self_loops: bool = False,
                            exclusive: bool = True,
                            limit_only_max_neighbours: bool = False,
                            numerical_tol: float = 1e-8,
                            manual_super_cell_radius: float = None,
                            super_cell_tol_factor: float = 0.25,
                            ) -> list:
    r"""Generate range connections for a primitive unit cell in a periodic lattice (vectorized).

    .. code-block:: python

        import numpy as np
        from kgcnn.graph.methods import range_neighbour_lattice

        artificial_lattice = np.array([[1.0, 0.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
        artificial_atoms = np.array([[0.1, 0.0, 0.0], [0.5, 0.5, 0.5]])
        out = range_neighbour_lattice(artificial_atoms, artificial_lattice)

        real_lattice = np.array([[-8.71172704, -0., -5.02971843],
                                 [-10.97279872, -0.01635133, 8.94600922],
                                 [-6.5538005, 12.48246168, 1.29207947]])
        real_atoms = np.array([[-24.14652308, 12.46611035, 6.41607351],
                               [-2.09180318, 0., -1.20770325],
                               [0., 0., 0.],
                               [-4.35586352, 0., -2.51485921]])
        out_real = range_neighbour_lattice(real_atoms, real_lattice)

    Args:
        coordinates (np.ndarray): Coordinate of nodes in the central primitive unit cell.
        lattice (np.ndarray): Lattice matrix of real space lattice vectors of shape `(3, 3)`.
            The lattice vectors must be given in rows of the matrix!
        max_distance (float, optional): Maximum distance to allow connections, can also be None. Defaults to 4.0.
        max_neighbours (int, optional): Maximum number of allowed neighbours for each central atom. Default is None.
        self_loops (bool, optional): Allow self-loops between the same central node. Defaults to False.
        exclusive (bool): Whether both distance and maximum neighbours must be fulfilled. Default is True.
        limit_only_max_neighbours (bool): Not used.
        numerical_tol  (float): Numerical tolerance for distance cut-off. Default is 1e-8.
        manual_super_cell_radius (float): Not used.
        super_cell_tol_factor (float): Tolerance to increase for search for neighbours. Default is 0.5.

    Returns:
        list: [indices, images, dist]
    """
    # Require either max_distance or max_neighbours to be specified.
    if max_distance is None and max_neighbours is None:
        raise ValueError("Need to specify either `max_distance` or `max_neighbours` or both.")

    # Volume and density of unit cell based on lattice matrix.
    volume_unit_cell = np.sum(np.abs(np.cross(lattice[0], lattice[1]) * lattice[2]))
    density_unit_cell = len(coordinates) / volume_unit_cell

    # Estimated real-space radius for max_neighbours based on density and volume of a single unit cell.
    if max_neighbours is not None:
        estimated_nn_volume = max_neighbours/density_unit_cell  # + len(coordinates)/ density_unit_cell
        estimated_nn_radius = abs(float(np.cbrt(estimated_nn_volume / np.pi * 3 / 4)))
        estimated_nn_radius = estimated_nn_radius
    else:
        estimated_nn_radius = None

    # Determine the required size of super-cell
    if manual_super_cell_radius is not None:
        if max_neighbours is None:
            # Does not make sense to specify manual supercell in this case.
            radius = max_distance
        else:
            radius = abs(manual_super_cell_radius)
    elif max_distance is None:
        radius = estimated_nn_radius
    elif max_neighbours is None:
        radius = max_distance
    else:
        if exclusive:
            radius = min(max_distance, estimated_nn_radius)
        else:
            radius = max(max_distance, estimated_nn_radius)

    _max_iter_nn_test = 100
    _iter_required = 0
    index1, index2, offset_vectors, distances = None, None, None, None
    for i in range(0, _max_iter_nn_test):
        _iter_required = i
        index1, index2, offset_vectors, distances = find_points_in_spheres(
            coordinates, coordinates,
            r=radius, pbc=np.array([True] * 3, dtype=int), lattice=lattice, tol=numerical_tol)
        offset_vectors = offset_vectors.astype('i2')

        # Remove self_loops:
        if not self_loops:
            no_self_loops = np.argwhere(~np.isclose(distances, 0)).reshape(-1)
            index1 = index1[no_self_loops]
            index2 = index2[no_self_loops]
            offset_vectors = offset_vectors[no_self_loops]
            distances = distances[no_self_loops]

        # We always sort here.
        def reorder(order, *args):
            return [x[order] for x in args]

        def limit_to(c, k, r, id1, id2, ov, dd, logical_reduce=None):
            splits = np.cumsum(c)[:-1]
            id1_split = np.split(id1, splits)
            id2_split = np.split(id2, splits)
            ov_split = np.split(ov, splits)
            dd_split = np.split(dd, splits)
            mask_r, mask_k = None, None
            if k is not None:
                mask_k = []
                for x in id1_split:
                    mask_per_node = np.zeros((len(x)), dtype="bool")
                    mask_per_node[:k] = True
                    mask_k.append(mask_per_node)
            if r is not None:
                mask_r = []
                for x in dd_split:
                    mask_per_node = x <= r + + abs(numerical_tol)
                    mask_r.append(mask_per_node)
            if k is not None and r is not None:
                mask = [logical_reduce(x1, x2) for x1, x2 in zip(mask_r, mask_k)]
            elif k is None:
                mask = mask_r
            else:
                mask = mask_k
            out_split = []
            for a in [id1_split, id2_split, ov_split, dd_split]:
                out_split.append(np.concatenate([x[mask[i_node]] for i_node, x in enumerate(a)], axis=0))
            return out_split

        sort_index = np.argsort(distances, kind="stable")
        index1, index2, offset_vectors, distances = reorder(sort_index, index1, index2, offset_vectors, distances)
        sort_index = np.argsort(index1, kind="stable")
        index1, index2, offset_vectors, distances = reorder(sort_index, index1, index2, offset_vectors, distances)

        # Case: radius cutoff. Only consider radius here.
        if max_neighbours is None:
            break

        unique_centers, counts = np.unique(index1, return_counts=True)
        enough_nn = not (np.any(counts < max_neighbours) if len(counts) > 0 else 0 < max_neighbours)
        # Case: knn.
        if max_distance is None:
            if not enough_nn:
                radius = radius * (1.0 + super_cell_tol_factor)
                continue
            else:
                index1, index2, offset_vectors, distances = limit_to(
                    counts, max_neighbours, None, index1, index2, offset_vectors, distances)
                break

        enough_distance = radius >= max_distance
        # Case mixed both radius and knn.
        if exclusive:
            if enough_nn or enough_distance:
                index1, index2, offset_vectors, distances = limit_to(
                    counts, max_neighbours, max_distance, index1, index2, offset_vectors, distances,
                    logical_reduce=np.logical_and)
                break
            else:
                radius = radius * (1.0 + super_cell_tol_factor)
                continue
        else:
            if not enough_nn or not enough_distance:
                radius = radius * (1.0 + super_cell_tol_factor)
                continue
            else:
                index1, index2, offset_vectors, distances = limit_to(
                    counts, max_neighbours, max_distance, index1, index2, offset_vectors, distances,
                    logical_reduce=np.logical_or)
                break

    if _iter_required + 1 >= _max_iter_nn_test:
        raise ValueError("Exceeded maximum number of allowed range extensions for neighbour calculation.")

    out_indices = np.concatenate([np.expand_dims(index1, axis=-1), np.expand_dims(index2, axis=-1)], axis=-1)
    return [out_indices, offset_vectors, distances]


# This is a python/numpy function to find neighbours in a periodic lattice.
# The pymatgen version is preferred, since this version can get slow for very skew lattice matrices.
def range_neighbour_lattice_python_vectorized(
        coordinates: np.ndarray, lattice: np.ndarray,
        max_distance: Union[float, None] = 4.0,
        max_neighbours: Union[int, None] = None,
        self_loops: bool = False,
        exclusive: bool = True,
        limit_only_max_neighbours: bool = False,
        numerical_tol: float = 1e-8,
        manual_super_cell_radius: float = None,
        super_cell_tol_factor: float = 0.25,
        ) -> list:
    r"""Generate range connections for a primitive unit cell in a periodic lattice (vectorized).

    The function generates a supercell of required radius and computes connections of neighbouring nodes
    from the primitive centered unit cell. For :obj:`max_neighbours` the supercell radius is estimated based on
    the unit cell density. Always the smallest necessary supercell is generated based on :obj:`max_distance` and
    :obj:`max_neighbours`. If a supercell for radius :obj:`max_distance` should always be generated but limited by
    :obj:`max_neighbours`, you can set :obj:`limit_only_max_neighbours` to `True`.

    .. warning::

        All atoms should be projected back into the primitive unit cell before calculating the range connections.

    .. note::

        For periodic structure, setting :obj:`max_distance` and :obj:`max_neighbours` to `inf` would also lead
        to an infinite number of neighbours and connections. If :obj:`exclusive` is set to `False`, having either
        :obj:`max_distance` or :obj:`max_neighbours` set to `inf`, will result in an infinite number of neighbours.
        If set to `None`, :obj:`max_distance` or :obj:`max_neighbours` can selectively be ignored.

    .. code-block:: python

        import numpy as np
        from kgcnn.graph.methods._periodic import range_neighbour_lattice_python_vectorized

        artificial_lattice = np.array([[1.0, 0.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
        artificial_atoms = np.array([[0.1, 0.0, 0.0], [0.5, 0.5, 0.5]])
        out = range_neighbour_lattice_python_vectorized(artificial_atoms, artificial_lattice)

        real_lattice = np.array([[-8.71172704, -0., -5.02971843],
                                 [-10.97279872, -0.01635133, 8.94600922],
                                 [-6.5538005, 12.48246168, 1.29207947]])
        real_atoms = np.array([[-24.14652308, 12.46611035, 6.41607351],
                               [-2.09180318, 0., -1.20770325],
                               [0., 0., 0.],
                               [-4.35586352, 0., -2.51485921]])
        out_real = range_neighbour_lattice_python_vectorized(real_atoms, real_lattice)

    Args:
        coordinates (np.ndarray): Coordinate of nodes in the central primitive unit cell.
        lattice (np.ndarray): Lattice matrix of real space lattice vectors of shape `(3, 3)`.
            The lattice vectors must be given in rows of the matrix!
        max_distance (float, optional): Maximum distance to allow connections, can also be None. Defaults to 4.0.
        max_neighbours (int, optional): Maximum number of allowed neighbours for each central atom. Default is None.
        self_loops (bool, optional): Allow self-loops between the same central node. Defaults to False.
        exclusive (bool): Whether both distance and maximum neighbours must be fulfilled. Default is True.
        limit_only_max_neighbours (bool): Whether to only use :obj:`max_neighbours` to limit the number of neighbours
            but not use it to calculate super-cell. Requires :obj:`max_distance` to be not `None`.
            Can be used if the super-cell should be generated with certain :obj:`max_distance`. Default is False.
        numerical_tol  (float): Numerical tolerance for distance cut-off. Default is 1e-8.
        manual_super_cell_radius (float): Manual radius for supercell. This is otherwise automatically set by either
            :obj:`max_distance` or :obj:`max_neighbours` or both. For manual supercell only. Default is None.
        super_cell_tol_factor (float): Tolerance factor for supercell relative to unit cell size. Default is 0.25.

    Returns:
        list: [indices, images, dist]
    """
    # Require either max_distance or max_neighbours to be specified.
    if max_distance is None and max_neighbours is None:
        raise ValueError("Need to specify either `max_distance` or `max_neighbours` or both.")

    # Here we set the lattice matrix, with lattice vectors in either columns or rows of the matrix.
    lattice_col = np.transpose(lattice)
    lattice_row = lattice

    # Index list for nodes. Enumerating the nodes in the central unit cell.
    node_index = np.expand_dims(np.arange(0, len(coordinates)), axis=1)  # Nx1

    # Diagonals, center, volume and density of unit cell based on lattice matrix.
    center_unit_cell = np.sum(lattice_row, axis=0, keepdims=True) / 2  # (1, 3)
    max_radius_cell = np.amax(np.sqrt(np.sum(np.square(lattice_row - center_unit_cell), axis=-1)))
    max_diameter_cell = 2*max_radius_cell
    volume_unit_cell = np.sum(np.abs(np.cross(lattice[0], lattice[1]) * lattice[2]))
    density_unit_cell = len(node_index) / volume_unit_cell

    # Center cell distance. Compute the distance matrix separately for the central primitive unit cell.
    # Here one can check if self-loops (meaning loops between the nodes of the central cell) should be allowed.
    center_indices = np.indices((len(node_index), len(node_index)))
    center_indices = center_indices.transpose(np.append(np.arange(1, 3), 0))  # NxNx2
    center_dist = np.expand_dims(coordinates, axis=0) - np.expand_dims(coordinates, axis=1)  # NxNx3
    center_image = np.zeros(center_dist.shape, dtype="int")
    if not self_loops:
        def remove_self_loops(x):
            m = np.logical_not(np.eye(len(x), dtype="bool"))
            x_shape = np.array(x.shape)
            x_shape[1] -= 1
            return np.reshape(x[m], x_shape)
        center_indices = remove_self_loops(center_indices)
        center_image = remove_self_loops(center_image)
        center_dist = remove_self_loops(center_dist)

    # Check the maximum atomic distance, since in practice atoms may not be inside the unit cell. Although they SHOULD
    # be projected back into the cell.
    max_diameter_atom_pair = np.amax(center_dist) if len(coordinates) > 1 else 0.0
    max_distance_atom_origin = np.amax(np.sqrt(np.sum(np.square(coordinates), axis=-1)))

    # Mesh Grid list. For a list of indices bounding left and right make a list of a 3D mesh.
    # Function is used to pad image unit cells or their origin for super-cell.
    def mesh_grid_list(bound_left: np.array, bound_right: np.array) -> np.array:
        pos = [np.arange(i, j+1, 1) for i, j in zip(bound_left, bound_right)]
        grid_list = np.array(np.meshgrid(*pos)).T.reshape(-1, 3)
        return grid_list

    # Estimated real-space radius for max_neighbours based on density and volume of a single unit cell.
    if max_neighbours is not None:
        estimated_nn_volume = (max_neighbours + len(node_index)) / density_unit_cell
        estimated_nn_radius = abs(float(np.cbrt(estimated_nn_volume / np.pi * 3 / 4)))
    else:
        estimated_nn_radius = None

    # Determine the required size of super-cell
    if manual_super_cell_radius is not None:
        super_cell_radius = abs(manual_super_cell_radius)
    elif max_distance is None:
        super_cell_radius = estimated_nn_radius
    elif max_neighbours is None or limit_only_max_neighbours:
        super_cell_radius = max_distance
    else:
        if exclusive:
            super_cell_radius = min(max_distance, estimated_nn_radius)
        else:
            super_cell_radius = max(max_distance, estimated_nn_radius)

    # Safety for super-cell radius. We add this distance to ensure that all atoms of the outer images are within the
    # actual cutoff distance requested.
    super_cell_tolerance = max(max_diameter_cell, max_diameter_atom_pair, max_distance_atom_origin)
    super_cell_tolerance *= (1.0 + super_cell_tol_factor)

    # Bounding box of real space cube with edge length 2 or inner sphere of radius 1 transformed into index
    # space gives 'bounding_box_unit'. Simply upscale for radius of super-cell.
    # To account for node pairing within the unit cell we add 'max_diameter_cell'.
    bounding_box_unit = np.sum(np.abs(np.linalg.inv(lattice_col)), axis=1)
    bounding_box_index = bounding_box_unit * (super_cell_radius + super_cell_tolerance)
    bounding_box_index = np.ceil(bounding_box_index).astype("int")

    # Making grid for super-cell that repeats the unit cell for required indices in 'bounding_box_index'.
    # Remove [0, 0, 0] of center unit cell by hand.
    bounding_grid = mesh_grid_list(-bounding_box_index, bounding_box_index)
    bounding_grid = bounding_grid[
        np.logical_not(np.all(bounding_grid == np.array([[0, 0, 0]]), axis=-1))]  # Remove center cell
    bounding_grid_real = np.dot(bounding_grid, lattice_row)

    # Check which centers are in the sphere of cutoff, since for non-rectangular lattice vectors, the parallelepiped
    # can be overshooting the required sphere. Better do this here, before computing coordinates of nodes.
    dist_centers = np.sqrt(np.sum(np.square(bounding_grid_real), axis=-1))
    mask_centers = dist_centers <= (super_cell_radius + super_cell_tolerance + abs(numerical_tol))
    images = bounding_grid[mask_centers]
    shifts = bounding_grid_real[mask_centers]

    # Compute node coordinates of images and prepare indices for those nodes. For 'N' nodes per cell and 'C' images
    # (without the central unit cell), this will be (flatten) arrays of (N*C)x3.
    num_images = images.shape[0]
    images = np.expand_dims(images, axis=0)  # 1xCx3
    images = np.repeat(images, len(coordinates), axis=0)  # NxCx3
    coord_images = np.expand_dims(coordinates, axis=1) + shifts  # NxCx3
    coord_images = np.reshape(coord_images, (-1, 3))  # (N*C)x3
    images = np.reshape(images, (-1, 3))  # (N*C)x3
    indices = np.expand_dims(np.repeat(node_index, num_images), axis=-1)  # (N*C)x1

    # Make distance matrix of central cell to all image. This will have shape Nx(NxC).
    dist = np.expand_dims(coord_images, axis=0) - np.expand_dims(coordinates, axis=1)  # Nx(N*C)x3
    dist_indices = np.concatenate(
        [np.repeat(np.expand_dims(node_index, axis=1), len(indices), axis=1),
         np.repeat(np.expand_dims(indices, axis=0), len(node_index), axis=0)], axis=-1)  # Nx(N*C)x2
    dist_images = np.repeat(np.expand_dims(images, axis=0), len(node_index), axis=0)  # Nx(N*C)x3

    # Adding distance matrix of nodes within the central cell to the image distance matrix.
    # The resulting shape is then Nx(NxC+1).
    dist_indices = np.concatenate([center_indices, dist_indices], axis=1)  # Nx(N*C+1)x2
    dist_images = np.concatenate([center_image, dist_images], axis=1)  # Nx(N*C+1)x2
    dist = np.concatenate([center_dist, dist], axis=1)  # Nx(N*C+1)x3

    # Distance in real space.
    dist = np.sqrt(np.sum(np.square(dist), axis=-1))  # Nx(N*C+1)

    # Sorting the distance matrix. Indices and image information must be sorted accordingly.
    arg_sort = np.argsort(dist, axis=-1)
    dist_sort = np.take_along_axis(dist, arg_sort, axis=1)
    dist_indices_sort = np.take_along_axis(
        dist_indices, np.repeat(np.expand_dims(arg_sort, axis=2), dist_indices.shape[2], axis=2), axis=1)
    dist_images_sort = np.take_along_axis(
        dist_images, np.repeat(np.expand_dims(arg_sort, axis=2), dist_images.shape[2], axis=2), axis=1)

    # Select range connections based on distance cutoff and nearest neighbour limit. Uses masking.
    # Based on 'max_distance'.
    mask_distance, mask_neighbours = None, None
    if max_distance is not None:
        mask_distance = dist_sort <= max_distance + abs(numerical_tol)
    # Based on 'max_neighbours'.
    if max_neighbours is not None:
        mask_neighbours = np.zeros_like(dist_sort, dtype="bool")
        mask_neighbours[:, :max_neighbours] = True

    if max_neighbours is None:
        mask = mask_distance
    elif max_distance is None:
        mask = mask_neighbours
    else:
        if exclusive:
            mask = np.logical_and(mask_neighbours, mask_distance)
        else:
            mask = np.logical_or(mask_neighbours, mask_distance)
    # Select nodes.
    out_dist = dist_sort[mask]
    out_images = dist_images_sort[mask]
    out_indices = dist_indices_sort[mask]

    return [out_indices, out_images, out_dist]



from keras.layers import Dense, Add, Multiply, Activation, Dropout, Subtract, Concatenate
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.aggr import AggregateLocalEdges
from kgcnn.layers.gather import GatherNodesOutgoing, GatherEdgesPairs
from kgcnn.layers.modules import Embedding
from kgcnn.layers.mlp import GraphMLP, MLP
from ._layers import PoolingNodesGRU


def model_disjoint(
        inputs,
        use_node_embedding=None,
        use_edge_embedding=None,
        input_node_embedding=None,
        input_edge_embedding=None,
        node_initialize=None,
        edge_initialize=None,
        depth=None,
        pooling_kwargs=None,
        edge_dense=None,
        edge_activation=None,
        dropout=None,
        node_dense=None,
        output_embedding=None,
        use_final_gru=None,
        output_mlp=None,
        pooling_gru=None
):
    n, ed, edi, e_pairs, batch_id_node, node_id, count_nodes = inputs
    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)

    h0 = Dense(**node_initialize)(n)
    he0 = Dense(**edge_initialize)(ed)

    # Model Loop
    h = h0
    he = he0
    for i in range(depth - 1):
        # Node message/update
        m_pool = AggregateLocalEdges(**pooling_kwargs)([h, he, edi])
        m_max = AggregateLocalEdges(pooling_method="max")([h, he, edi])
        m = Multiply()([m_pool, m_max])
        # In paper there is a potential COMMUNICATE() here but in reference code just add() operation.
        # Alternatives were GRU or MLP on concatenated messages.
        h = Add()([h, m])

        # Edge message/update
        h_out = GatherNodesOutgoing()([h, edi])
        e_rev = GatherEdgesPairs()([he, e_pairs])
        he = Subtract()([h_out, e_rev])
        he = Dense(**edge_dense)(he)
        he = Add()([he, he0])
        he = Activation(**edge_activation)(he)
        if dropout:
            he = Dropout(**dropout)(he)

    # Last step
    m_pool = AggregateLocalEdges(**pooling_kwargs)([h, he, edi])
    m_max = AggregateLocalEdges(pooling_method="max")([h, he, edi])
    m = Multiply()([m_pool, m_max])
    h_final = Concatenate()([m, h, h0])
    h_final = Dense(**node_dense)(h_final)

    n = h_final
    if output_embedding == 'graph':
        if use_final_gru:
            # Actually a simple GRU is not permutation invariant.
            # Better replace this by e.g. set2set or AttentivePooling.
            out = PoolingNodesGRU(**pooling_gru)([n, batch_id_node, node_id, count_nodes])
        else:
            out = PoolingNodes(**pooling_kwargs)([count_nodes, n, batch_id_node])
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported graph embedding for mode `CMPNN`")

    return out



from ._adj import (
    get_angle_indices, coordinates_to_distancematrix, invert_distance,
    define_adjacency_from_distance, sort_edge_indices, get_angle, add_edges_reverse_indices,
    rescale_edge_weights_degree_sym, add_self_loops_to_edge_indices, compute_reverse_edges_index_map,
    distance_to_gauss_basis, get_angle_between_edges, convert_scaled_adjacency_to_list
)
from ._geom import (
    get_principal_moments_of_inertia,
    shift_coordinates_to_unit_cell, distance_for_range_indices, distance_for_range_indices_periodic,
    coulomb_matrix_to_inverse_distance_proton, coordinates_from_distance_matrix
)
from ._periodic import (
    range_neighbour_lattice
)

__all__ = [
    # adj
    "get_angle_indices", "coordinates_to_distancematrix", "invert_distance",
    "define_adjacency_from_distance", "sort_edge_indices", "get_angle", "add_edges_reverse_indices",
    "rescale_edge_weights_degree_sym", "add_self_loops_to_edge_indices", "compute_reverse_edges_index_map",
    "distance_to_gauss_basis", "get_angle_between_edges", "convert_scaled_adjacency_to_list",
    # geom
    "get_principal_moments_of_inertia",
    "shift_coordinates_to_unit_cell", "distance_for_range_indices", "distance_for_range_indices_periodic",
    "coulomb_matrix_to_inverse_distance_proton", "coordinates_from_distance_matrix",
    # periodic
    "range_neighbour_lattice"
]



# Main package version.
__kgcnn_version__ = "4.0.1"

# Global definition of index order and axis.
__indices_axis__ = 0
__index_receive__ = 0
__index_send__ = 1

# Behaviour for backend functions.
__safe_scatter_max_min_to_zero__ = True

# Geometry
__geom_euclidean_norm_add_eps__ = True  # Set to false for exact sqrt computation for geometric layers.
__geom_euclidean_norm_no_nan__ = True  # Only used for inverse norm.



from keras.layers import Dense, Concatenate, Average, Activation
from kgcnn.layers.attention import AttentionHeadGAT
from kgcnn.layers.mlp import MLP, GraphMLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.pooling import PoolingNodes


def model_disjoint(inputs,
                   use_node_embedding: bool = None,
                   use_edge_embedding: bool = None,
                   input_node_embedding: dict = None,
                   input_edge_embedding: dict = None,
                   attention_args: dict = None,
                   pooling_nodes_args: dict = None,
                   depth: int = None,
                   attention_heads_num: int = None,
                   attention_heads_concat: bool = None,
                   output_embedding: str = None,
                   output_mlp: dict = None):
    # Model implementation with disjoint representation.
    n, ed, disjoint_indices, batch_id_node, count_nodes = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)

    # Model
    nk = Dense(units=attention_args["units"], activation="linear")(n)
    for i in range(0, depth):
        heads = [AttentionHeadGAT(**attention_args)([nk, ed, disjoint_indices]) for _ in range(attention_heads_num)]
        if attention_heads_concat:
            nk = Concatenate(axis=-1)(heads)
        else:
            nk = Average()(heads)
            nk = Activation(activation=attention_args["activation"])(nk)
    n = nk

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**pooling_nodes_args)([count_nodes, n, batch_id_node])
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for `GAT` .")

    return out



import keras as ks
from keras import ops


class CorrectPartialCharges(ks.layers.Layer):
    """Layer to compute average charge corrections for partial charges from total charge."""

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(CorrectPartialCharges, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        # Nothing to build here.
        self.built = True

    def call(self, inputs, **kwargs):
        r"""Computes charge correction.

        Args:
            inputs (list): [predict_tc, tot_charge, count_nodes, batch_id_node]

                - predict_tc (Tensor): Predicted total charge of shape `(batch, 1)`.
                - tot_charge (Tensor): True total charge of shape `(batch, 1)`.
                - count_nodes (Tensor): Number of nodes per sample of shape `(batch, )`.
                - batch_id_node (Tensor): Batch ID of nodes of shape `([N], )`.

        Returns:
            list: Corrections of partial charges of shape `([N], 1)`.
        """
        predict_tc, tot_charge, count_nodes, batch_id_node = inputs
        charge_diff = tot_charge-predict_tc
        avg_charge_diff = charge_diff/ops.expand_dims(count_nodes, axis=-1)
        avg_charge_diff_per_node = ops.take(avg_charge_diff, batch_id_node, axis=0)
        return avg_charge_diff_per_node



from keras.backend import backend
from typing import Union


def check_device():
    """Simple function to check for available devices for computing models. Mostly GPUs."""

    if backend() == "tensorflow":
        import tensorflow as tf
        cuda_is_available = tf.test.is_built_with_gpu_support() and tf.test.is_built_with_cuda()
        physical_device_list = tf.config.list_physical_devices()
        if physical_device_list:
            physical_device_name = [
                tf.config.experimental.get_device_details(x) for x in physical_device_list]
        else:
            physical_device_name = []
        logical_device_list = tf.config.experimental.list_logical_devices()
        if physical_device_name:
            try:
                memory_info = [tf.config.experimental.get_memory_info(x.name) for x in physical_device_list]
            except (TypeError, ValueError):
                memory_info = []
        else:
            memory_info = []

    elif backend() == "torch":
        import torch
        cuda_is_available = torch.cuda.is_available()
        physical_device_name = [torch.cuda.get_device_name(x) for x in range(torch.cuda.device_count())]
        logical_device_list = [x for x in range(torch.cuda.device_count())]
        memory_info = [{"allocated": round(torch.cuda.memory_allocated(i)/1024**3, 1),
                        "cached": round(torch.cuda.memory_reserved(i)/1024**3, 1)} for i in logical_device_list]
    elif backend() == "jax":
        import jax
        jax_devices = jax.devices()
        cuda_is_available = any([x.device_kind in ["gpu", "cuda"] for x in jax_devices])
        physical_device_name = [x for x in jax_devices]
        logical_device_list = [x.id for x in jax_devices]
        memory_info = [x.memory_stats() for x in jax_devices]

    else:
        raise NotImplementedError("Backend '%s' is not supported for `check_device_cuda` ." % backend())

    out_info = {
        "cuda_available": "%s" % cuda_is_available,
        "device_name": "%s" % physical_device_name,
        "device_id": "%s" % logical_device_list,
        "device_memory": "%s" % memory_info,
    }
    return out_info


def set_cuda_device(device_id: Union[int, list]):
    """Set the cuda device by ID.

    Better use cuda environment variable to do this instead of this function:

    .. code-block:: python

        import os
        os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"]="1"  # specify which GPU(s) to be used

    Args:
        device_id (int): ID of the GPU to set.
    """
    if backend() == "tensorflow":
        import tensorflow as tf
        gpus = tf.config.list_physical_devices('GPU')
        if isinstance(device_id, int):
            gpus_use = gpus[device_id]
        else:
            gpus_use = [gpus[i] for i in device_id]
        tf.config.set_visible_devices(gpus_use, 'GPU')
        if isinstance(gpus_use, list):
            for x in gpus_use:
                tf.config.experimental.set_memory_growth(x, True)
        else:
            tf.config.experimental.set_memory_growth(gpus_use, True)

    elif backend() == "torch":
        import torch
        torch.cuda.set_device(device_id)

    elif backend() == "jax":
        import jax
        jax.default_device = jax.devices('gpu')[device_id]

    else:
        raise NotImplementedError("Backend '%s' is not supported for `set_cuda_device` ." % backend())


import keras as ks
from typing import Union
from keras.layers import Layer, Concatenate
from keras import ops
from kgcnn import __indices_axis__ as global_axis_indices
from kgcnn import __index_send__ as global_index_send
from kgcnn import __index_receive__ as global_index_receive


class GatherNodes(Layer):
    r"""Gather node or edge embedding from an index list.

    The embeddings are gather from an index tensor. An edge is defined by index tuple :math:`(i ,j)` .
    In the default definition, index :math:`i` is expected to be the receiving or target node.
    Effectively, the layer simply does:

    .. code-block:: python

        ops.take(nodes, index[x], axis=0) for x in split_indices

    Additionally, the gathered embeddings can be concatenated along the index dimension,
    by setting :obj:`concat_axis` if index shape is known during build.

    .. note:

        Default of this layer is concatenation with :obj:`concat_axis=1` .

    Example of usage for :obj:`GatherNodes` :

    .. code-block:: python

        from keras import ops
        from kgcnn.layers.gather import GatherNodes
        nodes = ops.convert_to_tensor([[0.0],[1.0],[2.0],[3.0],[4.0]], dtype="float32")
        edge_idx = ops.convert_to_tensor([[0,0,1,2], [1,2,0,1]], dtype="int32")
        print(GatherNodes()([nodes, edge_idx]))

    """

    def __init__(self, split_indices=(0, 1),
                 concat_axis: Union[int, None] = 1,
                 axis_indices: int = global_axis_indices, **kwargs):
        """Initialize layer.

        Args:
            split_indices (list): List of indices to split and take values for. Default is (0, 1).
            concat_axis (int): The axis which concatenates embeddings. Default is 1.
            axis_indices (int): Axis on which to split indices from. Default is 0.
        """
        super(GatherNodes, self).__init__(**kwargs)
        self.split_indices = split_indices
        self.concat_axis = concat_axis
        self.axis_indices = axis_indices
        if self.concat_axis is not None:
            self._concat = Concatenate(axis=concat_axis)

    def _compute_gathered_shape(self, input_shape):
        assert len(input_shape) == 2
        x_shape, indices_shape = list(input_shape[0]), list(input_shape[1])
        xs = []
        indices_shape.pop(self.axis_indices)
        for _ in self.split_indices:
            xs.append(indices_shape + x_shape[1:])
        return xs

    def build(self, input_shape):
        """Build layer."""
        # We could call build on concatenate layer.
        xs = self._compute_gathered_shape(input_shape)
        if self.concat_axis is not None:
            self._concat.build(xs)
        self.built = True

    def compute_output_shape(self, input_shape):
        """Compute output shape of this layer."""
        xs = self._compute_gathered_shape(input_shape)
        if self.concat_axis is not None:
            xs = self._concat.compute_output_shape(xs)
        return xs

    def compute_output_spec(self, inputs_spec):
        output_shape = self.compute_output_shape([x.shape for x in inputs_spec])
        if self.concat_axis is not None:
            return ks.KerasTensor(output_shape, dtype=inputs_spec[0].dtype)
        return [ks.KerasTensor(s, dtype=inputs_spec[0].dtype) for s in output_shape]

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [nodes, index]

                - nodes (Tensor): Node embeddings of shape `([N], F)`
                - index (Tensor): Edge indices referring to nodes of shape `(2, [M])`

        Returns:
            Tensor: Gathered node embeddings that match the number of edges of shape `([M], 2*F)` or list of single
                node embeddings of shape [`([M], F)` , `([M], F)` , ...].
        """
        x, index = inputs
        gathered = []
        for i in self.split_indices:
            indices_take = ops.take(index, i, axis=self.axis_indices)
            x_i = ops.take(x, indices_take, axis=0)
            gathered.append(x_i)
        if self.concat_axis is not None:
            gathered = self._concat(gathered)
        return gathered

    def get_config(self):
        """Get config for this layer."""
        conf = super(GatherNodes, self).get_config()
        conf.update({"split_indices": self.split_indices, "concat_axis": self.concat_axis,
                     "axis_indices": self.axis_indices})
        return conf


class GatherNodesOutgoing(Layer):
    r"""Gather sending or outgoing nodes of edges with index :math:`j` .

    An edge is defined by index tuple :math:`(i, j)`.
    In the default definition, index :math:`j` is expected to be the sending or source node.
    """

    def __init__(self, selection_index: int = global_index_send, axis_indices: int = global_axis_indices, **kwargs):
        """Initialize layer.

        Args:
            selection_index (int): Index of sending nodes. Default is 1.
            axis_indices (int): Axis of node indices in index Tensor. Default is 0.
        """
        super(GatherNodesOutgoing, self).__init__(**kwargs)
        self.selection_index = selection_index
        self.axis_indices = axis_indices

    def build(self, input_shape):
        """Build layer."""
        super(GatherNodesOutgoing, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        """Compute output shape of this layer."""
        assert len(input_shape) == 2
        x_shape, indices_shape = list(input_shape[0]),  list(input_shape[1])
        indices_shape.pop(self.axis_indices)
        return tuple(indices_shape + x_shape[1:])

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [nodes, index]

                - nodes (Tensor): Node embeddings of shape `([N], F)`
                - index (Tensor): Edge indices referring to nodes of shape `(2, [M])`

        Returns:
            Tensor: Gathered node embeddings that match the number of edges of shape `([M], F)` .
        """
        x, index = inputs
        indices_take = ops.take(index, self.selection_index, axis=self.axis_indices)
        return ops.take(x, indices_take, axis=0)

    def get_config(self):
        """Get config for this layer."""
        conf = super(GatherNodesOutgoing, self).get_config()
        conf.update({"selection_index": self.selection_index, "axis_indices": self.axis_indices})
        return conf


class GatherNodesIngoing(Layer):
    r"""Gather receiving or ingoing nodes of edges with index :math:`i` .

    An edge is defined by index tuple :math:`(i, j)`.
    In the default definition, index :math:`i` is expected to be the receiving or target node.
    """

    def __init__(self, selection_index: int = global_index_receive, axis_indices: int = global_axis_indices, **kwargs):
        """Initialize layer.

        Args:
            selection_index (int): Index of receiving nodes. Default is 0.
            axis_indices (int): Axis of node indices in index Tensor. Default is 0.
        """
        super(GatherNodesIngoing, self).__init__(**kwargs)
        self.selection_index = selection_index
        self.axis_indices = axis_indices

    def build(self, input_shape):
        """Build layer."""
        super(GatherNodesIngoing, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        """Compute output shape of this layer."""
        assert len(input_shape) == 2
        x_shape, indices_shape = list(input_shape[0]),  list(input_shape[1])
        indices_shape.pop(self.axis_indices)
        return tuple(indices_shape + x_shape[1:])

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [nodes, index]

                - nodes (Tensor): Node embeddings of shape `([N], F)`
                - index (Tensor): Edge indices referring to nodes of shape `(2, [M])`

        Returns:
            Tensor: Gathered node embeddings that match the number of edges of shape `([M], F)` .
        """
        x, index = inputs
        indices_take = ops.take(index, self.selection_index, axis=self.axis_indices)
        return ops.take(x, indices_take, axis=0)

    def get_config(self):
        """Get config for this layer."""
        conf = super(GatherNodesIngoing, self).get_config()
        conf.update({"selection_index": self.selection_index, "axis_indices": self.axis_indices})
        return conf


class GatherState(Layer):
    r"""Layer to repeat environment or global state for a specific embeddings tensor like node or edge lists.

    To repeat the correct global state (like an environment feature vector) for each sub graph,
    a tensor with the target shape and batch ID is required.

    Mostly used to concatenate a global state :math:`\mathbf{s}` with node embeddings :math:`\mathbf{h}_i`
    like for example:

    .. math::

        \mathbf{h}_i = \mathbf{h}_i \oplus \mathbf{s}

    where this layer only repeats :math:`\mathbf{s}` to match an embedding tensor :math:`\mathbf{h}_i`.
    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(GatherState, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        super(GatherState, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        """Compute output shape of this layer."""
        assert len(input_shape) == 2
        state_shape, id_shape = list(input_shape[0]),  list(input_shape[1])
        return tuple(id_shape + state_shape[1:])

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [state, batch_id]

                - state (Tensor): Graph specific embedding tensor. This is tensor of shape `(batch, F)`
                - batch_id (Tensor): Tensor of batch IDs for each sub-graph of shape `([N], )` .

        Returns:
            Tensor: Graph embedding with repeated single state for each sub-graph of shape `([N], F)`.
        """
        env, batch_id = inputs
        out = ops.take(env, batch_id, axis=0)
        return out


class GatherEdgesPairs(Layer):
    """Gather edge pairs that also works for invalid indices given a certain pair, i.e. if an edge does not have its
    reverse counterpart in the edge indices list.

    This class is used in e.g. `DMPNN <https://pubs.acs.org/doi/full/10.1021/acs.jcim.9b00237>`__ .
    """

    def __init__(self, axis_indices: int = global_axis_indices, **kwargs):
        """Initialize layer.

        Args:
            axis_indices (int): Axis of indices. Default is 0.
        """
        super(GatherEdgesPairs, self).__init__(**kwargs)
        self.axis_indices = axis_indices

    def build(self, input_shape):
        """Build this layer."""
        self.built = True

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [edges, pair_index]

                - edges (Tensor): Edge embeddings of shape ([M], F)
                - pair_index (Tensor): Edge indices referring to edges of shape (1, [M])

        Returns:
            Tensor: Gathered edge embeddings that match the reverse edges of shape ([M], F) for index.
        """
        edges, pair_index = inputs
        indices_take = ops.take(pair_index, 0, self.axis_indices)
        index_corrected = ops.where(indices_take >= 0, indices_take, ops.zeros_like(indices_take))
        edges_paired = ops.take(edges, index_corrected, axis=0)
        edges_corrected = ops.where(
            ops.expand_dims(indices_take, axis=-1) >= 0, edges_paired, ops.zeros_like(edges_paired))
        return edges_corrected

    def get_config(self):
        """Get layer config."""
        conf = super(GatherEdgesPairs, self).get_config()
        conf.update({"axis_indices": self.axis_indices})
        return conf


import keras as ks
from keras.layers import Layer
from keras import ops
from keras import InputSpec
from kgcnn.ops.scatter import scatter_reduce_sum
from keras.layers import LayerNormalization
from keras.layers import BatchNormalization

global_normalization_args = {
    "GraphNormalization": (
        "mean_shift", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "alpha_initializer",
        "beta_regularizer", "gamma_regularizer", "beta_constraint", "alpha_constraint", "gamma_constraint",
        "alpha_regularizer"
    ),
    "GraphInstanceNormalization": (
        "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "alpha_initializer", "beta_regularizer",
        "gamma_regularizer", "beta_constraint", "alpha_constraint", "gamma_constraint", "alpha_regularizer"
    ),
    "GraphBatchNormalization": (
        "axis", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "beta_regularizer",
        "gamma_regularizer", "beta_constraint", "gamma_constraint", "momentum", "moving_mean_initializer",
        "moving_variance_initializer", "padded_disjoint"
    ),
    "GraphLayerNormalization": (
        "axis", "epsilon", "center", "scale", "beta_initializer", "gamma_initializer", "beta_regularizer",
        "gamma_regularizer", "beta_constraint", "gamma_constraint"
    ),
}


class GraphLayerNormalization(LayerNormalization):

    def __init__(self, **kwargs):
        super(GraphLayerNormalization, self).__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        return super(GraphLayerNormalization, self).compute_output_shape(input_shape[0])

    def build(self, input_shape):
        super(GraphLayerNormalization, self).build(input_shape[0])

    def call(self, inputs, **kwargs):
        return super(GraphLayerNormalization, self).call(inputs[0])

    def get_config(self):
        return super(GraphLayerNormalization, self).get_config()


class GraphBatchNormalization(BatchNormalization):

    def __init__(self, padded_disjoint: bool = False, **kwargs):
        super(GraphBatchNormalization, self).__init__(**kwargs)
        self.padded_disjoint = padded_disjoint

    def compute_output_shape(self, input_shape):
        return super(GraphBatchNormalization, self).compute_output_shape(input_shape[0])

    def build(self, input_shape):
        super(GraphBatchNormalization, self).build(input_shape[0])
        self.input_spec = [
            InputSpec(ndim=len(input_shape[0]), axes={self.axis: input_shape[0][self.axis]}),
            InputSpec(ndim=len(input_shape[1])),
            InputSpec(ndim=len(input_shape[2])),
        ]

    def call(self, inputs, training=None, **kwargs):
        if not self.padded_disjoint:
            return super(GraphBatchNormalization, self).call(inputs[0], training=training)
        else:
            padded_mask = inputs[1] > 0
            return super(GraphBatchNormalization, self).call(inputs[0], training=training, mask=padded_mask)

    def get_config(self):
        config = super(GraphBatchNormalization, self).get_config()
        config.update({"padded_disjoint": self.padded_disjoint})
        return config


class GraphNormalization(Layer):
    r"""Graph normalization for graph tensor objects.

    Following convention suggested by `GraphNorm: A Principled Approach (...) <https://arxiv.org/abs/2009.03294>`__ .

    The definition of normalization terms for graph neural networks can be categorized as follows. Here we copy the
    definition and description of `<https://arxiv.org/abs/2009.03294>`_ .

    .. math::

        \text{Norm}(\hat{h}_{i,j,g}) = \gamma \cdot \frac{\hat{h}_{i,j,g} - \mu}{\sigma} + \beta,


    Consider a batch of graphs :math:`{G_{1}, \dots , G_{b}}` where :math:`b` is the batch size.
    Let :math:`n_{g}` be the number of nodes in graph :math:`G_{g}` .
    We generally denote :math:`\hat{h}_{i,j,g}` as the inputs to the normalization module, e.g.,
    the :math:`j` -th feature value of node :math:`v_i` of graph :math:`G_{g}` ,
    :math:`i = 1, \dots , n_{g}` , :math:`j = 1, \dots , d` , :math:`g = 1, \dots , b` .

    For InstanceNorm, we regard each graph as an instance. The normalization is
    then applied to the feature values across all nodes for each
    individual graph, i.e., over dimension :math:`i` of :math:`\hat{h}_{i,j,g}` .

    Additionally, the following proposed additions for GraphNorm are added when compared to InstanceNorm.

    .. math::

        \text{GraphNorm}(\hat{h}_{i,j}) = \gamma_j \cdot \frac{\hat{h}_{i,j} - \alpha_j \mu_j }{\hat{\sigma}_j}+\beta_j

    where :math:`\mu_j = \frac{\sum^n_{i=1} \hat{h}_{i,j}}{n}` ,
    :math:`\hat{\sigma}^2_j = \frac{\sum^n_{i=1} (\hat{h}_{i,j} - \alpha_j \mu_j)^2}{n}`  ,
    and :math:`\gamma_j` , :math:`beta_j` are the affine parameters as in other normalization methods.

    .. code-block:: python

        from kgcnn.layers.norm import GraphNormalization
        layer = GraphNormalization()
    """

    def __init__(self,
                 mean_shift=True, epsilon=1e-3, center=True, scale=True,
                 beta_initializer='zeros', gamma_initializer='ones', alpha_initializer='ones',
                 beta_regularizer=None, gamma_regularizer=None, alpha_regularizer=None,
                 beta_constraint=None, gamma_constraint=None, alpha_constraint=None,
                 **kwargs):
        r"""Initialize layer :obj:`GraphBatchNormalization`.

        Args:
            epsilon: Small float added to variance to avoid dividing by zero. Defaults to 1e-3.
            center: If True, add offset of `beta` to normalized tensor. If False,
                `beta` is ignored. Defaults to True.
            scale: If True, multiply by `gamma`. If False, `gamma` is not used.
                Defaults to True. When the next layer is linear (also e.g. `nn.relu`),
                this can be disabled since the scaling will be done by the next layer.
            mean_shift (bool): Whether to apply alpha. Default is True.
            beta_initializer: Initializer for the beta weight. Defaults to 'zeros'.
            gamma_initializer: Initializer for the gamma weight. Defaults to 'ones'.
            alpha_initializer: Initializer for the alpha weight. Defaults to 'ones'.
            beta_regularizer: Optional regularizer for the beta weight. None by default.
            gamma_regularizer: Optional regularizer for the gamma weight. None by default.
            alpha_regularizer: Optional regularizer for the alpha weight. None by default.
            beta_constraint: Optional constraint for the beta weight. None by default.
            gamma_constraint: Optional constraint for the gamma weight. None by default.
            alpha_constraint: Optional constraint for the alpha weight. None by default.
        """
        super(GraphNormalization, self).__init__(**kwargs)
        self.epsilon = epsilon
        self._eps = ops.convert_to_tensor(epsilon, dtype=self.dtype)
        self.center = center
        self.mean_shift = mean_shift
        self.scale = scale
        self.beta_initializer = ks.initializers.get(beta_initializer)
        self.gamma_initializer = ks.initializers.get(gamma_initializer)
        self.alpha_initializer = ks.initializers.get(alpha_initializer)
        self.beta_regularizer = ks.regularizers.get(beta_regularizer)
        self.gamma_regularizer = ks.regularizers.get(gamma_regularizer)
        self.alpha_regularizer = ks.regularizers.get(alpha_regularizer)
        self.beta_constraint = ks.constraints.get(beta_constraint)
        self.gamma_constraint = ks.constraints.get(gamma_constraint)
        self.alpha_constraint = ks.constraints.get(alpha_constraint)
        # Weights
        self.alpha = None
        self.gamma = None
        self.beta = None

    def build(self, input_shape):
        param_shape = [x if x is not None else 1 for x in input_shape[0]]
        if self.scale:
            self.gamma = self.add_weight(
                name="gamma",
                shape=param_shape,
                initializer=self.gamma_initializer,
                regularizer=self.gamma_regularizer,
                constraint=self.gamma_constraint,
                trainable=True,
            )

        if self.center:
            self.beta = self.add_weight(
                name="beta",
                shape=param_shape,
                initializer=self.beta_initializer,
                regularizer=self.beta_regularizer,
                constraint=self.beta_constraint,
                trainable=True,
            )

        if self.mean_shift:
            self.alpha = self.add_weight(
                name="alpha",
                shape=param_shape,
                initializer=self.alpha_initializer,
                regularizer=self.alpha_regularizer,
                constraint=self.alpha_constraint,
                trainable=True,
            )

        self.built = True

    def _ragged_mean_std(self, inputs: list):
        values, row_ids, lengths = inputs

        if values.dtype in ("float16", "bfloat16") and self.dtype == "float32":
            values = ops.cast(values, "float32")

        shape_ = ops.shape(lengths)[:1] + ops.shape(values)[1:]
        counts_ = scatter_reduce_sum(row_ids, ops.ones_like(values), shape=shape_)

        mean = scatter_reduce_sum(row_ids, values, shape=shape_)/counts_
        if self.mean_shift:
            mean = mean * self.alpha
        mean = ops.take(mean, row_ids, axis=0)
        diff = values - mean
        # Not sure whether to stop gradients for variance if alpha ist used.
        square_diff = ops.square(diff)  # values - tf.stop_gradient(mean)
        variance = scatter_reduce_sum(row_ids, square_diff, shape=shape_)/counts_
        std = ops.sqrt(variance + self._eps)
        std = ops.take(std, row_ids, axis=0)
        return mean, std, diff / std

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): `[values, graph_id, reference]` .

                - values (Tensor): Tensor to normalize of shape `(None, F, ...)` .
                - graph_id (Tensor): Tensor of graph IDs of shape `(None, )` .
                - reference (Tensor, optional): Graph reference of disjoint batch of shape `(batch, )` .

        Returns:
            Tensor: Normalized tensor of identical shape (None, F, ...)
        """
        mean, std, new_values = self._ragged_mean_std(inputs)
        # Recomputing diff.
        if self.scale:
            new_values = new_values * self.gamma
        if self.center:
            new_values = new_values + self.beta
        return new_values

    def get_config(self):
        """Get layer configuration."""
        config = super(GraphNormalization, self).get_config()
        config.update({
            "mean_shift": self.mean_shift,
            "epsilon": self.epsilon,
            "center": self.center,
            "scale": self.scale,
            "beta_initializer": ks.initializers.serialize(self.beta_initializer),
            "gamma_initializer": ks.initializers.serialize(self.gamma_initializer),
            "alpha_initializer": ks.initializers.serialize(self.alpha_initializer),
            "beta_regularizer": ks.regularizers.serialize(self.beta_regularizer),
            "gamma_regularizer": ks.regularizers.serialize(self.gamma_regularizer),
            "alpha_regularizer": ks.regularizers.serialize(self.alpha_regularizer),
            "beta_constraint": ks.constraints.serialize(self.beta_constraint),
            "gamma_constraint": ks.constraints.serialize(self.gamma_constraint),
            "alpha_constraint": ks.constraints.serialize(self.alpha_constraint),
        })
        return config


class GraphInstanceNormalization(GraphNormalization):
    r"""Graph instance normalization for graph tensor objects.

    Following convention suggested by `GraphNorm: A Principled Approach (...) <https://arxiv.org/abs/2009.03294>`__ .

    The definition of normalization terms for graph neural networks can be categorized as follows. Here we copy the
    definition and description of `<https://arxiv.org/abs/2009.03294>`_ .

    .. math::

        \text{Norm}(\hat{h}_{i,j,g}) = \gamma \cdot \frac{\hat{h}_{i,j,g} - \mu}{\sigma} + \beta,

    Consider a batch of graphs :math:`{G_{1}, \dots , G_{b}}` where :math:`b` is the batch size.
    Let :math:`n_{g}` be the number of nodes in graph :math:`G_{g}` .
    We generally denote :math:`\hat{h}_{i,j,g}` as the inputs to the normalization module, e.g.,
    the :math:`j` -th feature value of node :math:`v_i` of graph :math:`G_{g}` ,
    :math:`i = 1, \dots , n_{g}` , :math:`j = 1, \dots , d` , :math:`g = 1, \dots , b` .

    For InstanceNorm, we regard each graph as an instance. The normalization is
    then applied to the feature values across all nodes for each
    individual graph, i.e., over dimension :math:`i` of :math:`\hat{h}_{i,j,g}` .

    .. code-block:: python

        from kgcnn.layers.norm import GraphInstanceNormalization
        layer = GraphInstanceNormalization()
    """

    def __init__(self, **kwargs):
        r"""Initialize layer :obj:`GraphBatchNormalization` .

        Args:
            epsilon: Small float added to variance to avoid dividing by zero. Defaults to 1e-3.
            center: If True, add offset of `beta` to normalized tensor. If False,
                `beta` is ignored. Defaults to True.
            scale: If True, multiply by `gamma`. If False, `gamma` is not used.
                Defaults to True. When the next layer is linear (also e.g. `nn.relu`),
                this can be disabled since the scaling will be done by the next layer.
            beta_initializer: Initializer for the beta weight. Defaults to 'zeros'.
            gamma_initializer: Initializer for the gamma weight. Defaults to 'ones'.
            alpha_initializer: Initializer for the alpha weight. Defaults to 'ones'.
            beta_regularizer: Optional regularizer for the beta weight. None by default.
            gamma_regularizer: Optional regularizer for the gamma weight. None by default.
            alpha_regularizer: Optional regularizer for the alpha weight. None by default.
            beta_constraint: Optional constraint for the beta weight. None by default.
            gamma_constraint: Optional constraint for the gamma weight. None by default.
            alpha_constraint: Optional constraint for the alpha weight. None by default.

        """
        super(GraphInstanceNormalization, self).__init__(mean_shift=False, **kwargs)



import keras as ks
from keras import ops
from keras.layers import Dense


class TrafoEdgeNetMessages(ks.layers.Layer):
    """Make message from edges for a linear transformation, i.e. matrix multiplication.
    The actual matrix is not a trainable weight of this layer but generated by a dense layer.
    This was proposed by `NMPNN <http://arxiv.org/abs/1704.01212>`__ .
    """

    def __init__(self, target_shape: tuple,
                 activation="linear",
                 use_bias=True,
                 kernel_regularizer=None, bias_regularizer=None, activity_regularizer=None,
                 kernel_constraint=None, bias_constraint=None,
                 kernel_initializer='glorot_uniform', bias_initializer='zeros',
                 **kwargs):
        """Initialize layer.

        Args:
            target_shape (tuple): Target shape for message matrix.
        """
        super(TrafoEdgeNetMessages, self).__init__(**kwargs)
        self.target_shape = target_shape
        self._units_out = int(target_shape[0])
        self._units_in = int(target_shape[1])
        self.lay_dense = Dense(units=self._units_out * self._units_in,
                               activation=activation, use_bias=use_bias,
                               kernel_regularizer=kernel_regularizer, bias_regularizer=bias_regularizer,
                               activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint,
                               bias_constraint=bias_constraint, kernel_initializer=kernel_initializer,
                               bias_initializer=bias_initializer)

    def build(self, input_shape):
        """Build layer."""
        super(TrafoEdgeNetMessages, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs (Tensor): Message embeddings or messages `([M], F)` .

        Returns:
            Tensor: Messages in matrix for multiplication of shape `([M], F_out, F_in)` .
        """
        up_scale = self.lay_dense(inputs, **kwargs)
        return ops.reshape(up_scale, (ops.shape(up_scale)[0], self._units_out, self._units_in))

    def get_config(self):
        """Update layer config."""
        config = super(TrafoEdgeNetMessages, self).get_config()
        config.update({"target_shape": self.target_shape})
        config_dense = self.lay_dense.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias"]:
            if x in config_dense.keys():
                config.update({x: config_dense[x]})
        return config



import keras as ks
import keras.metrics
import numpy as np
from keras import ops
import keras.saving
from kgcnn.ops.core import decompose_ragged_tensor


@ks.saving.register_keras_serializable(package='kgcnn', name='ScaledMeanAbsoluteError')
class ScaledMeanAbsoluteError(ks.metrics.MeanAbsoluteError):
    """Metric for a scaled mean absolute error (MAE), which can undo a pre-scaling of the targets. Only intended as
    metric this allows to info the MAE with correct units or absolute values during fit."""

    def __init__(self, scaling_shape=(), name='mean_absolute_error', dtype_scale: str = None, ragged: bool = False,
                 **kwargs):
        super(ScaledMeanAbsoluteError, self).__init__(name=name, **kwargs)
        self.scaling_shape = scaling_shape
        self._is_ragged = ragged
        self.dtype_scale = dtype_scale
        self.scale = self.add_variable(
            shape=scaling_shape,
            initializer=ks.initializers.Ones(),
            name='kgcnn_scale_mae',
            dtype=self.dtype_scale if self.dtype_scale is not None else self.dtype
        )

    def reset_state(self):
        for v in self.variables:
            if 'kgcnn_scale_mae' not in v.name:
                v.assign(ops.zeros(v.shape, dtype=v.dtype))

    def update_state(self, y_true, y_pred, sample_weight=None):
        if self._is_ragged:
            y_true = decompose_ragged_tensor(y_true)[0]
            y_pred = decompose_ragged_tensor(y_pred)[0]
        y_true = self.scale * ops.cast(y_true, dtype=self.scale.dtype)
        y_pred = self.scale * ops.cast(y_pred, dtype=self.scale.dtype)
        return super(ScaledMeanAbsoluteError, self).update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self):
        """Returns the serializable config of the metric."""
        conf = super(ScaledMeanAbsoluteError, self).get_config()
        conf.update({"scaling_shape": self.scaling_shape, "dtype_scale": self.dtype_scale,
                     "ragged": self._is_ragged})
        return conf

    def set_scale(self, scale):
        """Set the scale from numpy array. Usually used with broadcasting."""
        self.scale.assign(ops.cast(scale, dtype=scale.dtype))


@ks.saving.register_keras_serializable(package='kgcnn', name='ScaledRootMeanSquaredError')
class ScaledRootMeanSquaredError(ks.metrics.RootMeanSquaredError):
    """Metric for a scaled root mean squared error (RMSE), which can undo a pre-scaling of the targets.
    Only intended as metric this allows to info the MAE with correct units or absolute values during fit."""

    def __init__(self, scaling_shape=(), name='root_mean_squared_error', dtype_scale: str = None, ragged: bool = False,
                 **kwargs):
        super(ScaledRootMeanSquaredError, self).__init__(name=name, **kwargs)
        self.scaling_shape = scaling_shape
        self.dtype_scale = dtype_scale
        self._is_ragged = ragged
        self.scale = self.add_variable(
            shape=scaling_shape,
            initializer=ks.initializers.Ones(),
            name='kgcnn_scale_rmse',
            dtype=self.dtype_scale if self.dtype_scale is not None else self.dtype
        )

    def reset_state(self):
        for v in self.variables:
            if 'kgcnn_scale_rmse' not in v.name:
                v.assign(ops.zeros(v.shape, dtype=v.dtype))

    def update_state(self, y_true, y_pred, sample_weight=None):
        if self._is_ragged:
            y_true = decompose_ragged_tensor(y_true)[0]
            y_pred = decompose_ragged_tensor(y_pred)[0]
        y_true = self.scale * ops.cast(y_true, dtype=self.scale.dtype)
        y_pred = self.scale * ops.cast(y_pred, dtype=self.scale.dtype)
        return super(ScaledRootMeanSquaredError, self).update_state(y_true, y_pred, sample_weight=sample_weight)

    def get_config(self):
        """Returns the serializable config of the metric."""
        conf = super(ScaledRootMeanSquaredError, self).get_config()
        conf.update({"scaling_shape": self.scaling_shape, "dtype_scale": self.dtype_scale, "ragged": self._is_ragged})
        return conf

    def set_scale(self, scale):
        """Set the scale from numpy array. Usually used with broadcasting."""
        self.scale.assign(ops.cast(scale, dtype=scale.dtype))


@ks.saving.register_keras_serializable(package='kgcnn', name='ScaledMeanAbsoluteError')
class ScaledForceMeanAbsoluteError(ks.metrics.MeanMetricWrapper):
    """Metric for a scaled mean absolute error (MAE), which can undo a pre-scaling of the targets. Only intended as
    metric this allows to info the MAE with correct units or absolute values during fit."""

    def __init__(self, scaling_shape=(1, 1), name='force_mean_absolute_error', dtype_scale: str = None,
                 squeeze_states: bool = True, find_padded_atoms: bool = True, **kwargs):
        super(ScaledForceMeanAbsoluteError, self).__init__(fn=self.fn_force_mae, name=name, **kwargs)
        self.scaling_shape = scaling_shape
        self.dtype_scale = dtype_scale
        self.squeeze_states = squeeze_states
        self.find_padded_atoms = find_padded_atoms

        if scaling_shape[-1] == 1 and squeeze_states and len(scaling_shape) > 1:
            scaling_shape = scaling_shape[:-1]
        scaling_shape = tuple(list(scaling_shape[:1]) + [1, 1] + list(scaling_shape[1:]))

        self.scale = self.add_variable(
            shape=scaling_shape,
            initializer=ks.initializers.Ones(),
            name='kgcnn_scale_mae',
            dtype=self.dtype_scale if self.dtype_scale is not None else self.dtype
        )

    def fn_force_mae(self, y_true, y_pred):
        # (batch, N, 3)
        if self.find_padded_atoms:
            check_nonzero = ops.cast(ops.logical_not(
                ops.all(ops.isclose(y_true, ops.convert_to_tensor(0., dtype=y_true.dtype)), axis=2)), dtype="int32")
            row_count = ops.sum(check_nonzero, axis=1)
            row_count = ops.where(row_count < 1, 1, row_count)
            norm = 1/ops.cast(row_count, dtype=self.scale.dtype)
        else:
            norm = 1/ops.shape(y_true)[1]

        y_true = self.scale * ops.cast(y_true, dtype=self.scale.dtype)
        y_pred = self.scale * ops.cast(y_pred, dtype=self.scale.dtype)

        diff = ops.abs(y_true-y_pred)
        out = ops.sum(ops.mean(diff, axis=2), axis=1)*norm
        if not self.squeeze_states:
            out = ops.mean(out, axis=-1)
        return out

    def reset_state(self):
        for v in self.variables:
            if 'kgcnn_scale_mae' not in v.name:
                v.assign(ops.zeros(v.shape, dtype=v.dtype))

    def get_config(self):
        """Returns the serializable config of the metric."""
        # May not manage to deserialize `fn_force_mae`, set conf directly.
        # conf = super(ScaledForceMeanAbsoluteError, self).get_config()
        conf = {"name": self.name, "dtype": self.dtype}
        conf.update({"scaling_shape": self.scaling_shape, "dtype_scale": self.dtype_scale,
                     "find_padded_atoms": self.find_padded_atoms, "squeeze_states": self.squeeze_states})
        return conf

    def set_scale(self, scale):
        """Set the scale from numpy array. Usually used with broadcasting."""
        scaling_shape = scale.shape
        if scaling_shape[-1] == 1 and self.squeeze_states and len(scaling_shape) > 1:
            scale = np.squeeze(scale, axis=-1)
        scale = np.expand_dims(np.expand_dims(scale, axis=1), axis=2)
        self.scale.assign(ops.cast(scale, dtype=scale.dtype))


@ks.saving.register_keras_serializable(package='kgcnn', name='BinaryAccuracyNoNaN')
class BinaryAccuracyNoNaN(ks.metrics.MeanMetricWrapper):

    def __init__(self, name="binary_accuracy_no_nan",  dtype=None, threshold=0.5, **kwargs):
        if threshold is not None and (threshold <= 0 or threshold >= 1):
            raise ValueError(
                "Invalid value for argument `threshold`. "
                "Expected a value in interval (0, 1). "
                f"Received: threshold={threshold}"
            )
        super().__init__(
            fn=self._binary_accuracy_no_nan, name=name, dtype=dtype, threshold=threshold, **kwargs
        )
        self.threshold = threshold

    @staticmethod
    def _binary_accuracy_no_nan(y_true, y_pred, threshold=0.5):
        y_true = ops.convert_to_tensor(y_true)
        y_pred = ops.convert_to_tensor(y_pred)
        is_not_nan = ops.cast(ops.logical_not(ops.isnan(y_true)), y_true.dtype)
        threshold = ops.cast(threshold, y_pred.dtype)
        y_pred = ops.cast(y_pred > threshold, y_true.dtype)
        counts = ops.sum(ops.cast(
            ops.equal(y_true, y_pred), dtype=ks.backend.floatx()), axis=-1)
        norm = ops.sum(ops.cast(is_not_nan, dtype=ks.backend.floatx()), axis=-1)
        return counts/norm

    def get_config(self):
        config = {"name": self.name, "dtype": self.dtype, "threshold": self.threshold}
        return config


@ks.saving.register_keras_serializable(package='kgcnn', name='AUCNoNaN')
class AUCNoNaN(ks.metrics.AUC):

    def __init__(self, name="AUC_no_nan", **kwargs):
        super(AUCNoNaN, self).__init__(name=name, **kwargs)

    def update_state(self, y_true, y_pred, sample_weight=None):
        is_not_nan = ops.cast(ops.logical_not(ops.isnan(y_true)), y_true.dtype)
        if sample_weight is not None:
            sample_weight *= is_not_nan
        else:
            sample_weight = is_not_nan
        return super(AUCNoNaN, self).update_state(y_true, y_pred, sample_weight=sample_weight)


@ks.saving.register_keras_serializable(package='kgcnn', name='BalancedBinaryAccuracyNoNaN')
class BalancedBinaryAccuracyNoNaN(ks.metrics.SensitivityAtSpecificity):

    def __init__(self, name="balanced_binary_accuracy_no_nan", class_id=None, num_thresholds=1,
                 specificity=0.5, **kwargs):
        super(BalancedBinaryAccuracyNoNaN, self).__init__(name=name, class_id=class_id, num_thresholds=num_thresholds,
                                                          specificity=specificity, **kwargs)
        self._thresholds_distributed_evenly = False

    def update_state(self, y_true, y_pred, sample_weight=None):
        """Update the state of the metric.

        Args:
            y_true: Ground truth label values. shape = `[batch_size, d0, .. dN-1]`
                or shape = `[batch_size, d0, .. dN-1, 1]` .
            y_pred: The predicted probability values. shape = `[batch_size, d0, .. dN]` .
            sample_weight: Optional sample_weight acts as a coefficient for the metric.
        """
        is_not_nan = ops.cast(ops.logical_not(ops.isnan(y_true)), y_true.dtype)
        if sample_weight is not None:
            sample_weight *= is_not_nan
        else:
            sample_weight = is_not_nan

        return super(BalancedBinaryAccuracyNoNaN, self).update_state(
            y_true=y_true, y_pred=y_pred,
            sample_weight=sample_weight
        )

    def result(self):
        sensitivities = ops.divide(
            self.true_positives,
            self.true_positives + self.false_negatives + ks.config.epsilon(),
        )
        specificities = ops.divide(
            self.true_negatives,
            self.true_negatives + self.false_positives + ks.config.epsilon(),
        )
        result = (sensitivities + specificities)/2
        return result

    def get_config(self):
        config = super(BalancedBinaryAccuracyNoNaN, self).get_config()
        return config
